import logging
import os
from copy import deepcopy
from itertools import combinations

from palimpzest.constants import AggFunc, Model, PromptStrategy
from palimpzest.core.data.context_manager import ContextManager
from palimpzest.core.lib.schemas import (
    AUDIO_FIELD_TYPES,
    AUDIO_LIST_FIELD_TYPES,
    IMAGE_FIELD_TYPES,
    IMAGE_LIST_FIELD_TYPES,
)
from palimpzest.prompts import CONTEXT_SEARCH_PROMPT
from palimpzest.query.operators.aggregate import (
    ApplyGroupByOp,
    AverageAggregateOp,
    CountAggregateOp,
    MaxAggregateOp,
    MinAggregateOp,
    SemanticAggregate,
    SumAggregateOp,
)
from palimpzest.query.operators.compute import SmolAgentsCompute
from palimpzest.query.operators.convert import LLMConvertBonded, NonLLMConvert
from palimpzest.query.operators.critique_and_refine import CritiqueAndRefineConvert, CritiqueAndRefineFilter
from palimpzest.query.operators.distinct import DistinctOp
from palimpzest.query.operators.filter import LLMFilter, NonLLMFilter
from palimpzest.query.operators.join import EmbeddingJoin, NestedLoopsJoin, RelationalJoin
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.logical import (
    Aggregate,
    BaseScan,
    ComputeOperator,
    ContextScan,
    ConvertScan,
    Distinct,
    FilteredScan,
    GroupByAggregate,
    JoinOp,
    LimitScan,
    Project,
    SearchOperator,
    TopKScan,
)
from palimpzest.query.operators.mixture_of_agents import MixtureOfAgentsConvert, MixtureOfAgentsFilter
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.project import ProjectOp
from palimpzest.query.operators.rag import RAGConvert, RAGFilter
from palimpzest.query.operators.scan import ContextScanOp, MarshalAndScanDataOp
from palimpzest.query.operators.search import (
    SmolAgentsSearch,  # SmolAgentsCustomManagedSearch,  # SmolAgentsManagedSearch
)
from palimpzest.query.operators.split import SplitConvert, SplitFilter
from palimpzest.query.operators.topk import TopKOp
from palimpzest.query.optimizer.primitives import Expression, Group, LogicalExpression, PhysicalExpression
from palimpzest.utils.model_helpers import resolve_reasoning_settings

logger = logging.getLogger(__name__)


class Rule:
    """
    The abstract base class for transformation and implementation rules.
    """

    @classmethod
    def get_rule_id(cls):
        return cls.__name__

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        raise NotImplementedError("Calling this method from an abstract base class.")

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **kwargs: dict) -> set[Expression]:
        raise NotImplementedError("Calling this method from an abstract base class.")


class TransformationRule(Rule):
    """
    Base class for transformation rules which convert a logical expression to another logical expression.
    The substitute method for a TransformationRule should return all new expressions and all new groups
    which are created during the substitution.
    """

    @classmethod
    def is_exploration_rule(cls) -> bool:
        """Returns True if this rule is an exploration rule and False otherwise. Default is False."""
        return False

    @classmethod
    def substitute(
        cls, logical_expression: LogicalExpression, groups: dict[int, Group], expressions: dict[int, Expression], **kwargs
    ) -> tuple[set[LogicalExpression], set[Group]]:
        """
        This function applies the transformation rule to the logical expression, which
        potentially creates new intermediate expressions and groups.

        The function returns a tuple containing:
        - the set of all new logical expressions created when applying the transformation
        - the set of all new groups created when applying the transformation
        - the next group id (after creating any new groups)
        """
        raise NotImplementedError("Calling this method from an abstract base class.")


class ReorderConverts(TransformationRule):
    """
    This rule is an exploration rule that returns new logical expressions by re-ordering a sequence of ConvertScans.
    """

    @classmethod
    def is_exploration_rule(cls) -> bool:
        return True

    @classmethod
    def matches_pattern(cls, logical_expression: Expression) -> bool:
        is_match = isinstance(logical_expression.operator, ConvertScan)
        logger.debug(f"ReorderConverts matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(
        cls, logical_expression: LogicalExpression, groups: dict[int, Group], expressions: dict[int, Expression], **kwargs: dict
    ) -> tuple[set[LogicalExpression], set[Group]]:
        logger.debug(f"Substituting ReorderConverts for {logical_expression}")

        # initialize the sets of new logical expressions and groups to be returned
        new_logical_expressions, new_groups = set(), set()

        # for each input group, if this convert does not depend on an operator in that group:
        # then swap the group with this convert
        convert_operator: ConvertScan = logical_expression.operator
        for input_group_id in logical_expression.input_group_ids:
            input_group = groups[input_group_id]

            # if the convert's dependencies aren't contained within the input group's fields,
            # then we can not push it down into this group
            if any([field not in input_group.fields for field in convert_operator.depends_on]):
                continue

            # iterate over logical expressions
            for expr in input_group.logical_expressions:
                # if the expression operator is not a convert, we cannot swap
                if not isinstance(expr.operator, ConvertScan):
                    continue

                # if this convert depends on a field generated by the expression we're trying to swap with, we can't swap
                if any([field in expr.generated_fields for field in convert_operator.depends_on]):
                    continue

                # create new logical expression with convert pushed down to the input group's logical expression
                new_input_group_ids = deepcopy(expr.input_group_ids)
                new_input_fields = deepcopy(expr.input_fields)
                new_depends_on_field_names = deepcopy(logical_expression.depends_on_field_names)
                new_generated_fields = deepcopy(logical_expression.generated_fields)
                new_convert_expr = LogicalExpression(
                    convert_operator,
                    input_group_ids=new_input_group_ids,
                    input_fields=new_input_fields,
                    depends_on_field_names=new_depends_on_field_names,
                    generated_fields=new_generated_fields,
                    group_id=None,
                )

                # add new_convert_expr to set of new expressions
                new_logical_expressions.add(new_convert_expr)

                # get or compute the group_id and group for this new expression
                group_id, group = None, None

                # if the expression already exists, lookup the group_id and group
                if new_convert_expr.expr_id in expressions:
                    group_id = expressions[new_convert_expr.expr_id].group_id
                    new_convert_expr.set_group_id(group_id)
                    group = groups[group_id]

                # otherwise, lookup or create expression's group and add it to the new expressions
                else:
                    # first, compute the fields for the group
                    all_fields = {**new_input_fields, **new_generated_fields}

                    # next, compute the properties; the properties will be identical to those of the input group
                    # EXCEPT for the filters which will change as a result of our swap
                    new_group_properties = deepcopy(input_group.properties)

                    # if the expression we're swapping with is a map,
                    # we need to remove its model fields from the input group properties
                    if sorted(expr.operator.input_schema.model_fields.keys()) == sorted(expr.operator.output_schema.model_fields.keys()):
                        model_fields_dict = {
                            k: {"annotation": v.annotation, "default": v.default, "description": v.description}
                            for k, v in expr.operator.output_schema.model_fields.items()
                        }
                        new_group_properties["maps"].remove(model_fields_dict)

                    # finally, if this expression is a map, add its model fields to the new group's properties
                    if sorted(convert_operator.input_schema.model_fields.keys()) == sorted(convert_operator.output_schema.model_fields.keys()):
                        model_fields_dict = {
                            k: {"annotation": v.annotation, "default": v.default, "description": v.description}
                            for k, v in convert_operator.output_schema.model_fields.items()
                        }
                        if "maps" in new_group_properties:
                            new_group_properties["maps"].add(model_fields_dict)
                        else:
                            new_group_properties["maps"] = set([model_fields_dict])

                    # create group for this new convert expression
                    group = Group(
                        logical_expressions=[new_convert_expr],
                        fields=all_fields,
                        properties=new_group_properties,
                    )
                    group_id = group.group_id
                    new_convert_expr.set_group_id(group_id)

                    # if the group already exists, add the expression to that group
                    if group_id in groups:
                        group = groups[group_id]
                        group.logical_expressions.add(new_convert_expr)

                    # otherwise, add this new group to groups and to the set of new groups
                    else:
                        groups[group_id] = group
                        new_groups.add(group)

                # create final new logical expression with expr's operator pulled up
                new_expr = LogicalExpression(
                    expr.operator.copy(),
                    input_group_ids=[group_id] + [g_id for g_id in logical_expression.input_group_ids if g_id != input_group_id],
                    input_fields=group.fields,
                    depends_on_field_names=expr.depends_on_field_names,
                    generated_fields=expr.generated_fields,
                    group_id=logical_expression.group_id,
                )

                # add newly created expression to set of returned expressions
                new_logical_expressions.add(new_expr)

        logger.debug(f"Done substituting ReorderConverts for {logical_expression}")

        return new_logical_expressions, new_groups


class PushDownFilter(TransformationRule):
    """
    If this operator is a filter, push down the filter and replace it with the
    most expensive operator in the input group.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: Expression) -> bool:
        is_match = isinstance(logical_expression.operator, FilteredScan)
        logger.debug(f"PushDownFilter matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(
        cls, logical_expression: LogicalExpression, groups: dict[int, Group], expressions: dict[int, Expression], **kwargs: dict
    ) -> tuple[set[LogicalExpression], set[Group]]:
        logger.debug(f"Substituting PushDownFilter for {logical_expression}")

        # initialize the sets of new logical expressions and groups to be returned
        new_logical_expressions, new_groups = set(), set()

        # for each input group, if this filter does not depend on an operator
        # in that group: then swap the group with this filter
        filter_operator: FilteredScan = logical_expression.operator
        for input_group_id in logical_expression.input_group_ids:
            input_group = groups[input_group_id]

            # if the filter's dependencies aren't contained within the input group's fields,
            # then we can not push it down into this group
            if any([field not in input_group.fields for field in filter_operator.depends_on]):
                continue

            # iterate over logical expressions
            # NOTE: we previously deepcopy'ed the logical expression to avoid modifying the original;
            #       I think I've fixed this internally, but I'm leaving this NOTE as a reminder in case
            #       we see a regression / bug in the future
            for expr in input_group.logical_expressions:
                # if the expression operator is not a convert or a filter, we cannot swap
                if not (isinstance(expr.operator, (ConvertScan, FilteredScan, JoinOp))):
                    continue

                # if this filter depends on a field generated by the expression we're trying to swap with, we can't swap
                if any([field in expr.generated_fields for field in filter_operator.depends_on]):
                    continue

                # create new logical expression with filter pushed down to the input group's logical expression
                new_input_group_ids = deepcopy(expr.input_group_ids)
                new_input_fields = deepcopy(expr.input_fields)
                new_depends_on_field_names = deepcopy(logical_expression.depends_on_field_names)
                new_generated_fields = deepcopy(logical_expression.generated_fields)
                new_filter_expr = LogicalExpression(
                    filter_operator,
                    input_group_ids=new_input_group_ids,
                    input_fields=new_input_fields,
                    depends_on_field_names=new_depends_on_field_names,
                    generated_fields=new_generated_fields,
                    group_id=None,
                )

                # add new_filter_expr to set of new expressions
                new_logical_expressions.add(new_filter_expr)

                # get or compute the group_id and group for this new expression
                group_id, group = None, None

                # if the expression already exists, lookup the group_id and group
                if new_filter_expr.expr_id in expressions:
                    group_id = expressions[new_filter_expr.expr_id].group_id
                    new_filter_expr.set_group_id(group_id)
                    group = groups[group_id]

                # otherwise, lookup or create expression's group and add it to the new expressions
                else:
                    # first, compute the fields for the group
                    all_fields = {**new_input_fields, **new_generated_fields}

                    # next, compute the properties; the properties will be identical to those of the input group
                    # EXCEPT for the filters which will change as a result of our swap
                    new_group_properties = deepcopy(input_group.properties)

                    # if the expression we're swapping with is a FilteredScan,
                    # we need to remove its filter from the input group properties
                    if isinstance(expr.operator, FilteredScan):
                        filter_str = expr.operator.filter.get_filter_str()
                        new_group_properties["filters"].remove(filter_str)

                    # finally, add the pushed-down filter to the new group's properties
                    filter_str = filter_operator.filter.get_filter_str()
                    if "filters" in new_group_properties:
                        new_group_properties["filters"].add(filter_str)
                    else:
                        new_group_properties["filters"] = set([filter_str])

                    # create group for this new filter expression
                    group = Group(
                        logical_expressions=[new_filter_expr],
                        fields=all_fields,
                        properties=new_group_properties,
                    )
                    group_id = group.group_id
                    new_filter_expr.set_group_id(group_id)

                    # if the group already exists, add the expression to that group
                    if group_id in groups:
                        group = groups[group_id]
                        group.logical_expressions.add(new_filter_expr)

                    # otherwise, add this new group to groups and to the set of new groups
                    else:
                        groups[group_id] = group
                        new_groups.add(group)

                # create final new logical expression with expr's operator pulled up
                new_expr = LogicalExpression(
                    expr.operator.copy(),
                    input_group_ids=[group_id] + [g_id for g_id in logical_expression.input_group_ids if g_id != input_group_id],
                    input_fields=group.fields,
                    depends_on_field_names=expr.depends_on_field_names,
                    generated_fields=expr.generated_fields,
                    group_id=logical_expression.group_id,
                )

                # add newly created expression to set of returned expressions
                new_logical_expressions.add(new_expr)

        logger.debug(f"Done substituting PushDownFilter for {logical_expression}")

        return new_logical_expressions, new_groups


class ImplementationRule(Rule):
    """
    Base class for implementation rules which convert a logical expression to a physical expression.
    """

    @classmethod
    def _get_image_fields(cls, logical_expression: LogicalExpression) -> set[str]:
        """Returns the set of fields which have an image (or list[image]) type."""
        return set([
            field_name.split(".")[-1]
            for field_name, field in logical_expression.input_fields.items()
            if field.annotation in IMAGE_FIELD_TYPES and field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _get_list_image_fields(cls, logical_expression: LogicalExpression) -> set[str]:
        """Returns the set of fields which have a list[image] type."""
        return set([
            field_name.split(".")[-1]
            for field_name, field in logical_expression.input_fields.items()
            if field.annotation in IMAGE_LIST_FIELD_TYPES and field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _get_audio_fields(cls, logical_expression: LogicalExpression) -> set[str]:
        """Returns the set of fields which have an audio (or list[audio]) type."""
        return set([
            field_name.split(".")[-1]
            for field_name, field in logical_expression.input_fields.items()
            if field.annotation in AUDIO_FIELD_TYPES and field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _get_list_audio_fields(cls, logical_expression: LogicalExpression) -> set[str]:
        """Returns the set of fields which have a list[audio] type."""
        return set([
            field_name.split(".")[-1]
            for field_name, field in logical_expression.input_fields.items()
            if field.annotation in AUDIO_LIST_FIELD_TYPES and field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_image_only_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes only image input(s) and False otherwise."""
        return all([
            field.annotation in IMAGE_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_image_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes image input(s) and False otherwise."""
        return any([
            field.annotation in IMAGE_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_audio_only_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes only audio input(s) and False otherwise."""
        return all([
            field.annotation in AUDIO_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_audio_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes audio input(s) and False otherwise."""
        return any([
            field.annotation in AUDIO_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_text_only_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes only text input(s) and False otherwise."""
        return all([
            field.annotation not in IMAGE_FIELD_TYPES + AUDIO_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    @classmethod
    def _is_text_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes text input(s) and False otherwise."""
        return any([
            field.annotation not in IMAGE_FIELD_TYPES + AUDIO_FIELD_TYPES
            for field_name, field in logical_expression.input_fields.items()
            if field_name.split(".")[-1] in logical_expression.depends_on_field_names
        ])

    # TODO: support powerset of text + image + audio (+ video) multi-modal operations
    @classmethod
    def _is_text_image_multimodal_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes text and image inputs and False otherwise."""
        return cls._is_image_operation(logical_expression) and cls._is_text_operation(logical_expression)

    @classmethod
    def _is_text_audio_multimodal_operation(cls, logical_expression: LogicalExpression) -> bool:
        """Returns True if the logical_expression processes text and audio inputs and False otherwise."""
        return cls._is_audio_operation(logical_expression) and cls._is_text_operation(logical_expression)

    @classmethod
    def _model_matches_input(cls, model: Model, logical_expression: LogicalExpression) -> bool:
        """Returns True if the model is capable of processing the input and False otherwise."""
        # compute how many image fields are in the input, and whether any fields are list[image] fields
        num_image_fields = len(cls._get_image_fields(logical_expression))
        has_list_image_field = len(cls._get_list_image_fields(logical_expression)) > 0
        num_audio_fields = len(cls._get_audio_fields(logical_expression))
        has_list_audio_field = len(cls._get_list_audio_fields(logical_expression)) > 0

        # corner-case: for now, all operators use text or vision models for processing inputs to __call__
        if model.is_embedding_model():
            return False

        # corner-case: Llama vision models cannot handle multiple image inputs (at least using Together)
        if model.is_llama_model() and model.is_vision_model() and (num_image_fields > 1 or has_list_image_field):
            return False

        # corner-case: Gemini models cannot handle multiple audio inputs
        if model.is_vertex_model() and model.is_audio_model() and (num_audio_fields > 1 or has_list_audio_field):
            return False

        # text-only input and text supporting model
        if cls._is_text_only_operation(logical_expression) and model.is_text_model():
            return True

        # image-only input and image supporting model
        if cls._is_image_only_operation(logical_expression) and model.is_vision_model():
            return True

        # audio-only input and audio supporting model
        if cls._is_audio_only_operation(logical_expression) and model.is_audio_model():
            return True

        # multi-modal input and multi-modal supporting model
        if cls._is_text_image_multimodal_operation(logical_expression) and model.is_text_image_multimodal_model():  # noqa: SIM103
            return True

        # multi-modal input and multi-modal supporting model
        if cls._is_text_audio_multimodal_operation(logical_expression) and model.is_text_audio_multimodal_model():  # noqa: SIM103
            return True

        return False

    @classmethod
    def _get_fixed_op_kwargs(cls, logical_expression: LogicalExpression, runtime_kwargs: dict) -> dict:
        """Get the fixed set of physical op kwargs provided by the logical expression and the runtime keyword arguments."""
        # get logical operator 
        logical_op = logical_expression.operator

        # set initial set of parameters for physical op
        op_kwargs = logical_op.get_logical_op_params()
        op_kwargs.update(
            {
                "verbose": runtime_kwargs["verbose"],
                "logical_op_id": logical_op.get_logical_op_id(),
                "unique_logical_op_id": logical_op.get_unique_logical_op_id(),
                "logical_op_name": logical_op.logical_op_name(),
                "api_base": runtime_kwargs["api_base"],
            }
        )

        return op_kwargs

    @classmethod
    def _perform_substitution(
        cls,
        logical_expression: LogicalExpression,
        physical_op_class: type[PhysicalOperator],
        runtime_kwargs: dict,
        variable_op_kwargs: list[dict] | dict | None = None,
    ) -> set[PhysicalExpression]:
        """
        This performs basic substitution logic which proceeds in four steps:

            1. The basic kwargs for the physical operator are computed using the logical operator
               and runtime kwargs.
            2. If variable kwargs are provided, then they are merged with the basic kwargs and one
               instance of the physical operator is created for each dictionary of variable kwargs.
            3. A physical expression is created for each physical operator instance.
            4. The unique set of physical expressions is returned.

        Args:
            logical_expression (LogicalExpression): The logical expression containing a logical operator.
            physical_op_class (type[PhysicalOperator]): The class of the physical operator we wish to construct.
            runtime_kwargs (dict): Keyword arguments which are provided at runtime.
            variable_op_kwargs (list[dict] | dict | None): A (list of) variable kwargs to customize each
                physical operator instance.

        Returns:
            set[PhysicalExpression]: The unique set of physical expressions produced by initializing the
                physical_op_class with the provided keyword arguments.
        """
        # get physical operator kwargs which are fixed for each instance of the physical operator
        fixed_op_kwargs = cls._get_fixed_op_kwargs(logical_expression, runtime_kwargs)

        # make variable_op_kwargs a list of dictionaries
        if variable_op_kwargs is None:
            variable_op_kwargs = [{}]
        elif isinstance(variable_op_kwargs, dict):
            variable_op_kwargs = [variable_op_kwargs]

        # construct physical operators for each set of kwargs
        physical_expressions = []
        for var_op_kwargs in variable_op_kwargs:
            # get kwargs for this physical operator instance
            op_kwargs = {**fixed_op_kwargs, **var_op_kwargs}

            # construct the physical operator
            op = physical_op_class(**op_kwargs)

            # construct physical expression and add to list of expressions
            expression = PhysicalExpression.from_op_and_logical_expr(op, logical_expression)
            physical_expressions.append(expression)

        return set(physical_expressions)


class NonLLMConvertRule(ImplementationRule):
    """
    Substitute a logical expression for a UDF ConvertScan with a NonLLMConvert physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, ConvertScan) and logical_expression.operator.udf is not None
        logger.debug(f"NonLLMConvertRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting NonLLMConvertRule for {logical_expression}")
        return cls._perform_substitution(logical_expression, NonLLMConvert, runtime_kwargs)


class LLMConvertBondedRule(ImplementationRule):
    """
    Substitute a logical expression for a ConvertScan with a bonded convert physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, ConvertScan) and logical_expression.operator.udf is None
        logger.debug(f"LLMConvertBondedRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting LLMConvertBondedRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = PromptStrategy.MAP if use_reasoning_prompt else PromptStrategy.MAP_NO_REASONING
            variable_op_kwargs.append(
                {
                    "model": model,
                    "prompt_strategy": prompt_strategy,
                    "reasoning_effort": reasoning_effort,
                }
            )

        return cls._perform_substitution(logical_expression, LLMConvertBonded, runtime_kwargs, variable_op_kwargs)


class RAGRule(ImplementationRule):
    """
    Implementation rule for the RAG operators.
    """

    num_chunks_per_fields = [1, 2, 4]
    chunk_sizes = [1000, 2000, 4000]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_map_match = isinstance(logical_op, ConvertScan) and cls._is_text_only_operation(logical_expression) and logical_op.udf is None
        is_filter_match = isinstance(logical_op, FilteredScan) and cls._is_text_only_operation(logical_expression) and logical_op.filter.filter_fn is None
        logger.debug(f"RAGRule matches_pattern: {is_map_match or is_filter_match} for {logical_expression}")
        return is_map_match or is_filter_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting RAGRule for {logical_expression}")
        # select physical operator class based on whether this is a map or filter operation
        phys_op_cls = RAGConvert if isinstance(logical_expression.operator, ConvertScan) else RAGFilter

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
             use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
             prompt_strategy = (
                 PromptStrategy.MAP if use_reasoning_prompt else PromptStrategy.MAP_NO_REASONING
                 if phys_op_cls is RAGConvert
                 else PromptStrategy.FILTER if use_reasoning_prompt else PromptStrategy.FILTER_NO_REASONING
             )
             variable_op_kwargs.extend(
                 [
                    {
                        "model": model,
                        "prompt_strategy": prompt_strategy,
                        "num_chunks_per_field": num_chunks_per_field,
                        "chunk_size": chunk_size,
                        "reasoning_effort": reasoning_effort,
                    }
                    for num_chunks_per_field in cls.num_chunks_per_fields
                    for chunk_size in cls.chunk_sizes
                 ]
             )

        return cls._perform_substitution(logical_expression, phys_op_cls, runtime_kwargs, variable_op_kwargs)


class MixtureOfAgentsRule(ImplementationRule):
    """
    Implementation rule for the MixtureOfAgents operators.
    """

    num_proposer_models = [1, 2, 3]
    temperatures = [0.0, 0.4, 0.8]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_map_match = isinstance(logical_op, ConvertScan) and logical_op.udf is None
        is_filter_match = isinstance(logical_op, FilteredScan) and logical_op.filter.filter_fn is None
        logger.debug(f"MixtureOfAgentsRule matches_pattern: {is_map_match or is_filter_match} for {logical_expression}")
        return is_map_match or is_filter_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting MixtureOfAgentsRule for {logical_expression}")
        # select physical operator class based on whether this is a map or filter operation
        phys_op_cls = MixtureOfAgentsConvert if isinstance(logical_expression.operator, ConvertScan) else MixtureOfAgentsFilter

        # create variable physical operator kwargs for each model which can implement this logical_expression
        _, reasoning_effort = resolve_reasoning_settings(None, runtime_kwargs["reasoning_effort"])
        proposer_model_set = {model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)}
        aggregator_model_set = {model for model in runtime_kwargs["available_models"] if model.is_text_model()}
        variable_op_kwargs = [
            {
                "proposer_models": list(proposer_models),
                "temperatures": [temp] * len(proposer_models),
                "aggregator_model": aggregator_model,
                "reasoning_effort": reasoning_effort,
            }
            for k in cls.num_proposer_models
            for temp in cls.temperatures
            for proposer_models in combinations(proposer_model_set, k)
            for aggregator_model in aggregator_model_set
        ]

        return cls._perform_substitution(logical_expression, phys_op_cls, runtime_kwargs, variable_op_kwargs)


class CritiqueAndRefineRule(ImplementationRule):
    """
    Implementation rule for the CritiqueAndRefine operators.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_map_match = isinstance(logical_op, ConvertScan) and logical_op.udf is None
        is_filter_match = isinstance(logical_op, FilteredScan) and logical_op.filter.filter_fn is None
        logger.debug(f"CritiqueAndRefineRule matches_pattern: {is_map_match or is_filter_match} for {logical_expression}")
        return is_map_match or is_filter_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting CritiqueAndRefineRule for {logical_expression}")
        # select physical operator class based on whether this is a map or filter operation
        phys_op_cls = CritiqueAndRefineConvert if isinstance(logical_expression.operator, ConvertScan) else CritiqueAndRefineFilter

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = (
                PromptStrategy.MAP if use_reasoning_prompt else PromptStrategy.MAP_NO_REASONING
                if phys_op_cls is CritiqueAndRefineConvert
                else PromptStrategy.FILTER if use_reasoning_prompt else PromptStrategy.FILTER_NO_REASONING
            )
            variable_op_kwargs.extend(
                [
                    {
                        "model": model,
                        "critic_model": critic_model,
                        "refine_model": refine_model,
                        "prompt_strategy": prompt_strategy,
                        "reasoning_effort": reasoning_effort,
                    }
                    for critic_model in models
                    for refine_model in models
                ]
            )

        return cls._perform_substitution(logical_expression, phys_op_cls, runtime_kwargs, variable_op_kwargs)


class SplitRule(ImplementationRule):
    """
    Implementation rule for the Split operators.
    """
    num_chunks = [2, 4, 6]
    min_size_to_chunk = [1000, 4000]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_map_match = isinstance(logical_op, ConvertScan) and cls._is_text_only_operation(logical_expression) and logical_op.udf is None
        is_filter_match = isinstance(logical_op, FilteredScan) and cls._is_text_only_operation(logical_expression) and logical_op.filter.filter_fn is None
        logger.debug(f"SplitRule matches_pattern: {is_map_match or is_filter_match} for {logical_expression}")
        return is_map_match or is_filter_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting SplitRule for {logical_expression}")
        # select physical operator class based on whether this is a map or filter operation
        phys_op_cls = SplitConvert if isinstance(logical_expression.operator, ConvertScan) else SplitFilter

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        _, reasoning_effort = resolve_reasoning_settings(None, runtime_kwargs["reasoning_effort"])
        variable_op_kwargs = [
            {
                "model": model,
                "min_size_to_chunk": min_size_to_chunk,
                "num_chunks": num_chunks,
                "reasoning_effort": reasoning_effort,
            }
            for model in models
            for min_size_to_chunk in cls.min_size_to_chunk
            for num_chunks in cls.num_chunks
        ]

        return cls._perform_substitution(logical_expression, phys_op_cls, runtime_kwargs, variable_op_kwargs)


class TopKRule(ImplementationRule):
    """
    Substitute a logical expression for a TopKScan with a TopK physical implementation.
    """
    k_budgets = [1, 3, 5, 10, 15, 20, 25]

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, TopKScan)
        logger.debug(f"TopKRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting TopKRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        ks = cls.k_budgets if logical_expression.operator.k == -1 else [logical_expression.operator.k]
        variable_op_kwargs = [{"k": k} for k in ks]
        return cls._perform_substitution(logical_expression, TopKOp, runtime_kwargs, variable_op_kwargs)


class NonLLMFilterRule(ImplementationRule):
    """
    Substitute a logical expression for a FilteredScan with a non-llm filter physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_match = isinstance(logical_op, FilteredScan) and logical_op.filter.filter_fn is not None
        logger.debug(f"NonLLMFilterRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting NonLLMFilterRule for {logical_expression}")
        return cls._perform_substitution(logical_expression, NonLLMFilter, runtime_kwargs)


class LLMFilterRule(ImplementationRule):
    """
    Substitute a logical expression for a FilteredScan with an llm filter physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op = logical_expression.operator
        is_match = isinstance(logical_op, FilteredScan) and logical_op.filter.filter_fn is None
        logger.debug(f"LLMFilterRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting LLMFilterRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = PromptStrategy.FILTER if use_reasoning_prompt else PromptStrategy.FILTER_NO_REASONING
            variable_op_kwargs.append(
                {
                    "model": model,
                    "prompt_strategy": prompt_strategy,
                    "reasoning_effort": reasoning_effort,
                }
            )

        return cls._perform_substitution(logical_expression, LLMFilter, runtime_kwargs, variable_op_kwargs)


class RelationalJoinRule(ImplementationRule):
    """
    Substitute a logical expression for a JoinOp with a RelationalJoin physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, JoinOp) and logical_expression.operator.condition == ""
        logger.debug(f"RelationalJoinRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting RelationalJoinRule for {logical_expression}")
        return cls._perform_substitution(logical_expression, RelationalJoin, runtime_kwargs)


class NestedLoopsJoinRule(ImplementationRule):
    """
    Substitute a logical expression for a JoinOp with an (LLM) NestedLoopsJoin physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, JoinOp) and logical_expression.operator.condition != ""
        logger.debug(f"NestedLoopsJoinRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting NestedLoopsJoinRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = PromptStrategy.JOIN if use_reasoning_prompt else PromptStrategy.JOIN_NO_REASONING
            variable_op_kwargs.append(
                {
                    "model": model,
                    "prompt_strategy": prompt_strategy,
                    "join_parallelism": runtime_kwargs["join_parallelism"],
                    "reasoning_effort": reasoning_effort,
                    "retain_inputs": not runtime_kwargs["is_validation"],
                }
            )

        return cls._perform_substitution(logical_expression, NestedLoopsJoin, runtime_kwargs, variable_op_kwargs)


class EmbeddingJoinRule(ImplementationRule):
    """
    Substitute a logical expression for a JoinOp with an EmbeddingJoin physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, JoinOp) and logical_expression.operator.condition != "" and not cls._is_audio_operation(logical_expression)
        logger.debug(f"EmbeddingJoinRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting EmbeddingJoinRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression)]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = PromptStrategy.JOIN if use_reasoning_prompt else PromptStrategy.JOIN_NO_REASONING
            variable_op_kwargs.append(
                {
                    "model": model,
                    "prompt_strategy": prompt_strategy,
                    "join_parallelism": runtime_kwargs["join_parallelism"],
                    "reasoning_effort": reasoning_effort,
                    "retain_inputs": not runtime_kwargs["is_validation"],
                    "num_samples": 10, # TODO: iterate over different choices of num_samples
                }
            )

        return cls._perform_substitution(logical_expression, EmbeddingJoin, runtime_kwargs, variable_op_kwargs)

class SemanticAggregateRule(ImplementationRule):
    """
    Substitute a logical expression for a SemanticAggregate with an llm physical implementation.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, Aggregate) and logical_expression.operator.agg_str is not None
        logger.debug(f"SemanticAggregateRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting SemanticAggregateRule for {logical_expression}")

        # create variable physical operator kwargs for each model which can implement this logical_expression
        models = [model for model in runtime_kwargs["available_models"] if cls._model_matches_input(model, logical_expression) and not model.is_llama_model()]
        variable_op_kwargs = []
        for model in models:
            use_reasoning_prompt, reasoning_effort = resolve_reasoning_settings(model, runtime_kwargs["reasoning_effort"])
            prompt_strategy = PromptStrategy.AGG if use_reasoning_prompt else PromptStrategy.AGG_NO_REASONING
            variable_op_kwargs.append(
                {
                    "model": model,
                    "prompt_strategy": prompt_strategy,
                    "reasoning_effort": reasoning_effort,
                }
            )

        return cls._perform_substitution(logical_expression, SemanticAggregate, runtime_kwargs, variable_op_kwargs)


class AggregateRule(ImplementationRule):
    """
    Substitute the logical expression for an aggregate with its physical counterpart.
    """

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, Aggregate) and logical_expression.operator.agg_func is not None
        logger.debug(f"AggregateRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting AggregateRule for {logical_expression}")

        # get the physical op class based on the aggregation function
        physical_op_class = None
        if logical_expression.operator.agg_func == AggFunc.COUNT:
            physical_op_class = CountAggregateOp
        elif logical_expression.operator.agg_func == AggFunc.AVERAGE:
            physical_op_class = AverageAggregateOp
        elif logical_expression.operator.agg_func == AggFunc.SUM:
            physical_op_class = SumAggregateOp
        elif logical_expression.operator.agg_func == AggFunc.MIN:
            physical_op_class = MinAggregateOp
        elif logical_expression.operator.agg_func == AggFunc.MAX:
            physical_op_class = MaxAggregateOp
        else:
            raise Exception(f"Cannot support aggregate function: {logical_expression.operator.agg_func}")

        # perform the substitution
        return cls._perform_substitution(logical_expression, physical_op_class, runtime_kwargs)


class AddContextsBeforeComputeRule(ImplementationRule):
    """
    Searches the ContextManager for additional contexts which may be useful for the given computation.

    TODO: track cost of generating search query
    """
    k = 1
    SEARCH_GENERATOR_PROMPT = CONTEXT_SEARCH_PROMPT

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        is_match = isinstance(logical_expression.operator, ComputeOperator)
        logger.debug(f"AddContextsBeforeComputeRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting AddContextsBeforeComputeRule for {logical_expression}")

        # load an LLM to generate a short search query
        model = None
        if os.getenv("OPENAI_API_KEY"):
            model = "openai/gpt-4o-mini"
        elif os.getenv("ANTHROPIC_API_KEY"):
            model = "anthropic/claude-3-5-sonnet-20241022"
        elif os.getenv("GEMINI_API_KEY"):
            model = "vertex_ai/gemini-2.0-flash"
        elif os.getenv("TOGETHER_API_KEY"):
            model = "together_ai/meta-llama/Llama-3.3-70B-Instruct-Turbo"

        # importing litellm here because importing above causes deprecation warning
        import litellm

        # retrieve any additional context which may be useful
        cm = ContextManager()
        response = litellm.completion(
            model=model,
            messages=[{"role": "user", "content": cls.SEARCH_GENERATOR_PROMPT.format(instruction=logical_expression.operator.instruction)}]
        )
        query = response.choices[0].message.content
        variable_op_kwargs = {"additional_contexts": cm.search_context(query, k=cls.k, where={"materialized": True})}
        return cls._perform_substitution(logical_expression, SmolAgentsCompute, runtime_kwargs, variable_op_kwargs)


class BasicSubstitutionRule(ImplementationRule):
    """
    For logical operators with a single physical implementation, substitute the
    logical expression with its physical counterpart.
    """

    LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP = {
        BaseScan: MarshalAndScanDataOp,
        # ComputeOperator: SmolAgentsCompute,
        SearchOperator: SmolAgentsSearch, # SmolAgentsManagedSearch, # SmolAgentsCustomManagedSearch
        ContextScan: ContextScanOp,
        Distinct: DistinctOp,
        LimitScan: LimitScanOp,
        Project: ProjectOp,
        GroupByAggregate: ApplyGroupByOp,
    }

    @classmethod
    def matches_pattern(cls, logical_expression: LogicalExpression) -> bool:
        logical_op_class = logical_expression.operator.__class__
        is_match = logical_op_class in cls.LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP
        logger.debug(f"BasicSubstitutionRule matches_pattern: {is_match} for {logical_expression}")
        return is_match

    @classmethod
    def substitute(cls, logical_expression: LogicalExpression, **runtime_kwargs) -> set[PhysicalExpression]:
        logger.debug(f"Substituting BasicSubstitutionRule for {logical_expression}")
        physical_op_class = cls.LOGICAL_OP_CLASS_TO_PHYSICAL_OP_CLASS_MAP[logical_expression.operator.__class__]
        return cls._perform_substitution(logical_expression, physical_op_class, runtime_kwargs)



from __future__ import annotations

import base64
import os
from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import Path

import pandas as pd
from bs4 import BeautifulSoup
from pydantic import BaseModel

from palimpzest import constants
from palimpzest.core.data import dataset
from palimpzest.core.lib.schemas import (
    AudioFile,
    DefaultSchema,
    ImageFile,
    PDFFile,
    TextFile,
    WebPage,
    XLSFile,
    create_schema_from_df,
    create_schema_from_fields,
)
from palimpzest.query.operators.logical import BaseScan
from palimpzest.tools.pdfparser import get_text_from_pdf


####################
### BASE CLASSES ###
####################
class IterDataset(dataset.Dataset, ABC):
    """
    The `IterDataset` is an abstract base class for root `Datasets` whose data is accessed
    via iteration. Classes which inherit from this class must implement two methods:

    - `__len__()`: which returns the number of elements in the dataset
    - `__getitem__(idx: int)`: which takes in an `idx` and returns the element at that index
    """

    def __init__(self, id: str, schema: type[BaseModel] | list[dict]) -> None:
        """
            Constructor for the `IterDataset` class.

            Args:
                id (str): a string identifier for the `Dataset`
                schema (BaseModel | list[dict]): The output schema of the records returned by the `Dataset`
        """
        # compute Schema and call parent constructor
        schema = create_schema_from_fields(schema) if isinstance(schema, list) else schema
        super().__init__(sources=None, operator=BaseScan(datasource=self, output_schema=schema), schema=schema, id=id)

    @abstractmethod
    def __len__(self) -> int:
        """Returns the number of items in the `Dataset`."""
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> dict:
        """
        Returns a single item from the `Dataset` at the given index.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary representing the item at the given index. The dictionary
                  keys (i.e. fields) should match the fields specified in the schema of the
                  dataset, and the values should be the values associated with those fields.

                    # Example return value
                    {"field1": value1, "field2": value2, ...}

        """
        pass


class BaseFileDataset(IterDataset):
    """
    BaseFileDataset is the base class for multiple `IterDatasets` which iterate over
    different types of files.
    """

    def __init__(self, path: str, **kwargs) -> None:
        """
        Constructor for the `BaseFileDataset` class.

        Args:
            path (str): The path to the file
            kwargs (dict): Keyword arguments containing the `Dataset's` id and file-specific `Schema`
        """
        # check that path is a valid file or directory
        assert os.path.isfile(path) or os.path.isdir(path), f"Path {path} is not a file nor a directory"

        # get list of filepaths
        self.filepaths = []
        if os.path.isfile(path):
            self.filepaths = [path]
        else:
            self.filepaths = [
                os.path.join(path, filename)
                for filename in sorted(os.listdir(path))
                if os.path.isfile(os.path.join(path, filename))
            ]

        # call parent constructor to set id, operator, and schema
        super().__init__(**kwargs)

    def __len__(self) -> int:
        return len(self.filepaths)


class BaseFileDirectoryDataset(IterDataset):
    """
    BaseFileDirectoryDataset is the base class for multiple `IterDatasets` which iterate over
    different types of files. This class walks the entire directory tree rooted at `path`.
    """

    def __init__(self, path: str, **kwargs) -> None:
        """
        Constructor for the `BaseFileDataset` class.

        Args:
            path (str): The path to the file
            kwargs (dict): Keyword arguments containing the `Dataset's` id and file-specific `Schema`
        """
        # check that path is a valid file or directory
        assert os.path.isfile(path) or os.path.isdir(path), f"Path {path} is not a file nor a directory"

        # get list of filepaths
        self.filepaths = []
        if os.path.isfile(path):
            self.filepaths = [path]
        else:
            self.filepaths = []
            for root, _, files in os.walk(path):
                for file in files:
                    fp = os.path.join(root, file)
                    self.filepaths.append(fp)
            self.filepaths = sorted(self.filepaths)

        # call parent constructor to set id, operator, and schema
        super().__init__(**kwargs)

    def __len__(self) -> int:
        return len(self.filepaths)

########################
### CONCRETE CLASSES ###
########################
class MemoryDataset(IterDataset):
    """
    MemoryDataset returns one or more dictionaries that reflect the contents of an in-memory Python object `vals`.
    If `vals` is not a pd.DataFrame, then the dictionary returned by `__getitem__()` has a single field called "value".
    Otherwise, the dictionary contains the key-value mapping from columns to values for the `idx` row in the dataframe.

    TODO(gerardo): Add support for other types of in-memory data structures (he has some code for subclassing
        MemoryDataset on his branch)
    """

    def __init__(self, id: str, vals: list | pd.DataFrame, schema: type[BaseModel] | list[dict] | None = None) -> None:
        """
        Constructor for the `MemoryDataset` class. The `schema` is set to the default `DefaultSchema` schema.
        If `vals` is a pd.DataFrame, then the schema is set to the schema inferred from the DataFrame.

        Args:
            id (str): a string identifier for the `Dataset`
            vals (Any): The in-memory data to iterate over
        """
        # if list[dict] --> convert to pd.DataFrame first
        self.vals = pd.DataFrame(vals) if isinstance(vals, list) and all([isinstance(item, dict) for item in vals]) else vals
        if schema is None:
            schema = create_schema_from_df(self.vals) if isinstance(self.vals, pd.DataFrame) else DefaultSchema
        super().__init__(id=id, schema=schema)

    def __len__(self) -> int:
        return len(self.vals)

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the value(s) for the element at the specified `idx` in `vals`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: If `vals` is not a pd.DataFrame, then the dictionary has a single field called "value".
                Otherwise, the dictionary contains the key-value mapping from columns to values for the
                `idx` row in the dataframe.
            
            .. code-block:: python
            
                # Example return value at idx = 0, for the following list of values
                # [42, 43, 44, ...]
                {"value": 42}

                # Example return value at idx = 0, for the following DataFrame:
                # +---------+---------+---------+
                # |  name   |   job   |  hobby  |
                # +---------+---------+---------+
                # |  Alice  |  doctor |  tennis |
                # |  Bob    |  lawyer |  chess  |
                # +---------+---------+---------+
                {"name": "Alice", "job": "doctor", "hobby": "tennis"}
        """
        item = (
            self.vals.iloc[idx].to_dict()
            if isinstance(self.vals, pd.DataFrame)
            else {"value": self.vals[idx]}
        )

        return item


class HTMLFileDataset(BaseFileDataset):
    """
    HTMLFileDataset returns a dictionary for each HTML file in a directory. Each dictionary contains the
    filename, raw HTML content, and parsed content of a single HTML file in the directory.
    """
    def __init__(self, id: str, path: str) -> None:
        """
        Constructor for the `HTMLFileDataset` class. The `schema` is set to the `WebPage` schema.

        Args:
            id (str): a string identifier for the `Dataset`
            path (str): The path to the directory
        """
        super().__init__(path=path, id=id, schema=WebPage)
        self.filepaths = [fp for fp in self.filepaths if fp.endswith(tuple(constants.HTML_EXTENSIONS))]

    def _html_to_text_with_links(self, html: str) -> str:
        # Parse the HTML content
        soup = BeautifulSoup(html, "html.parser")

        # Find all hyperlink tags
        for a in soup.find_all("a"):
            # Check if the hyperlink tag has an 'href' attribute
            if a.has_attr("href"):
                # Replace the hyperlink with its text and URL in parentheses
                a.replace_with(f"{a.text} ({a['href']})")

        # Extract text from the modified HTML
        text = soup.get_text(separator="\n", strip=True)

        return text

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, raw HTML content, and parsed content of the HTML file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return
        
        Returns:
            dict: A dictionary with the filename, raw HTML content, and parsed content of the HTML file.

            .. code-block:: python

                {
                    "filename": "file.html",
                    "html": "raw HTML content here",
                    "text": "parsed text content here",
                }
        """
        item = {}
        filepath = self.filepaths[idx]
        item["filename"] = os.path.basename(filepath)
        with open(filepath) as f:
            text_content = f.read()

        html = text_content
        tokens = html.split()[: constants.MAX_HTML_ROWS]
        item["html"] = " ".join(tokens)

        stripped_html = self._html_to_text_with_links(text_content)
        tokens = stripped_html.split()[: constants.MAX_HTML_ROWS]
        item["text"] = " ".join(tokens)

        return item


class ImageFileDataset(BaseFileDataset):
    """
    ImageFileDataset returns a dictionary for each image file in a directory. Each dictionary contains the
    filename and the base64 encoded bytes content of a single image file in the directory.
    """
    def __init__(self, id: str, path: str) -> None:
        """
        Constructor for the `ImageFileDataset` class. The `schema` is set to the `ImageFile` schema.

        Args:
            id (str): a string identifier for the `Dataset`
            path (str): The path to the directory
        """
        super().__init__(path=path, id=id, schema=ImageFile)
        self.filepaths = [fp for fp in self.filepaths if fp.endswith(tuple(constants.IMAGE_EXTENSIONS))]

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and base64 encoded bytes content of the image file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename and base64 encoded bytes content of the image file.

            .. code-block:: python

                {
                    "filename": "image.jpg",
                    "contents": b"base64 encoded image content here",
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            contents = base64.b64encode(f.read()).decode("utf-8")

        return {"filename": filename, "contents": contents}


class PDFFileDataset(BaseFileDataset):
    """
    PDFFileDataset returns a dictionary for each PDF file in a directory. Each dictionary contains the
    filename, raw PDF content, and parsed text content of a single PDF file in the directory.

    This class also uses one of a predefined set of PDF processors to extract text content from the PDF files.
    """
    def __init__(
        self,
        id: str,
        path: str,
        pdfprocessor: str = "pypdf",
        file_cache_dir: str = "/tmp",
    ) -> None:
        """
        Constructor for the `PDFFileDataset` class. The `schema` is set to the `PDFFile` schema.

        Args:
            id (str): a string identifier for the `Dataset`
            path (str): The path to the directory
            pdfprocessor (str): The PDF processor to use for extracting text content from the PDF files
            file_cache_dir (str): The directory to store the temporary files generated during PDF processing
        """
        super().__init__(path=path, id=id, schema=PDFFile)
        self.filepaths = [fp for fp in self.filepaths if fp.endswith(tuple(constants.PDF_EXTENSIONS))]
        self.pdfprocessor = pdfprocessor
        self.file_cache_dir = file_cache_dir

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, raw PDF content, and parsed text content of the PDF file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename, raw PDF content, and parsed text content of the PDF file.

            .. code-block:: python

                {
                    "filename": "file.pdf",
                    "contents": b"raw PDF content here",
                    "text_contents": "parsed text content here",
                }
        """
        filepath = self.filepaths[idx]
        pdf_filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            pdf_bytes = f.read()

        # generate text_content from PDF
        text_content = get_text_from_pdf(pdf_filename, pdf_bytes, pdfprocessor=self.pdfprocessor, file_cache_dir=self.file_cache_dir)

        # construct and return item
        return {"filename": pdf_filename, "contents": pdf_bytes, "text_contents": text_content}


class TextFileDataset(BaseFileDataset):
    """
    TextFileDataset returns a dictionary for each text file in a directory. Each dictionary contains the
    filename and contents of a single text file in the directory.
    """
    def __init__(self, id: str, path: str) -> None:
        """
        Constructor for the `TextFileDataset` class. The `schema` is set to the `TextFile` schema.

        Args:
            id (str): a string identifier for the `Dataset`
            path (str): The path to the directory
        """
        super().__init__(path=path, id=id, schema=TextFile)

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and contents of the text file at the specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename and contents of the text file.

            .. code-block:: python

                {
                    "filename": "file.txt",
                    "contents": "text content here",
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath) as f:
            contents = f.read()

        return {"filename": filename, "contents": contents}


class XLSFileDataset(BaseFileDataset):
    """
    XLSFileDataset returns a dictionary for each XLS file in a directory. Each dictionary contains the
    filename, contents, sheet names, and the number of sheets for a single XLS file in the directory.
    """
    def __init__(self, id: str, path: str) -> None:
        """
        Constructor for the `XLSFileDataset` class. The `schema` is set to the `XLSFile` schema.
        """
        super().__init__(path=path, id=id, schema=XLSFile)
        self.filepaths = [fp for fp in self.filepaths if fp.endswith(tuple(constants.XLS_EXTENSIONS))]

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, contents, sheet names, and the number of sheets of the XLS file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename, contents, sheet names, and the number of sheets of the XLS file.

            .. code-block:: python

                {
                    "filename": "file.xls",
                    "contents": b"raw XLS content here",
                    "sheet_names": ["Sheet1", "Sheet2", "Sheet3],
                    "number_sheets": 3,
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            contents = f.read()

        xls = pd.ExcelFile(BytesIO(contents), engine="openpyxl")

        return {
            "filename": filename,
            "contents": contents,
            "sheet_names": xls.sheet_names,
            "number_sheets": len(xls.sheet_names),
        }


class AudioFileDataset(BaseFileDirectoryDataset):
    """
    AudioFileDataset returns a dictionary for each audio file in a directory. Each dictionary contains the
    filename and the base64 encoded bytes content of a single audio file in the directory.
    """
    def __init__(self, id: str, path: str) -> None:
        """
        Constructor for the `AudioFileDataset` class. The `schema` is set to the `AudioFile` schema.

        Args:
            id (str): a string identifier for the `Dataset`
            path (str): The path to the directory
        """
        super().__init__(path=path, id=id, schema=AudioFile)
        self.filepaths = [fp for fp in self.filepaths if fp.endswith(tuple(constants.AUDIO_EXTENSIONS))]

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and base64 encoded bytes content of the audio file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename and base64 encoded bytes content of the audio file.

            .. code-block:: python

                {
                    "filename": "audio.wav",
                    "contents": b"base64 encoded audio content here",
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            contents = base64.b64encode(f.read()).decode("utf-8")

        return {"filename": filename, "contents": contents}


def get_local_source(id: str, path: str | Path, **kwargs) -> dataset.Dataset:
    """Return a `Dataset` for a local file or directory."""
    if os.path.isfile(path):
        return TextFileDataset(id, path)

    elif os.path.isdir(path):
        if all([f.endswith(tuple(constants.IMAGE_EXTENSIONS)) for f in os.listdir(path)]):
            return ImageFileDataset(id, path)

        elif all([f.endswith(tuple(constants.PDF_EXTENSIONS)) for f in os.listdir(path)]):
            pdfprocessor = kwargs.get("pdfprocessor", constants.DEFAULT_PDF_PROCESSOR)
            file_cache_dir = kwargs.get("file_cache_dir", "/tmp")
            return PDFFileDataset(
                id=id, path=path, pdfprocessor=pdfprocessor, file_cache_dir=file_cache_dir
            )

        elif all([f.endswith(tuple(constants.XLS_EXTENSIONS)) for f in os.listdir(path)]):
            return XLSFileDataset(id, path)

        elif all([f.endswith(tuple(constants.HTML_EXTENSIONS)) for f in os.listdir(path)]):
            return HTMLFileDataset(id, path)

        else:
            return TextFileDataset(id, path)
    else:
        raise ValueError(f"Path {path} is invalid. Does not point to a file or directory.")


def resolve_datasource(id: str, source: str | Path | list | pd.DataFrame, **kwargs) -> dataset.Dataset:
    """
    This helper function returns a `Dataset` object based on the `source` type.
    The returned `Dataset` object is guaranteed to have a schema.
    """
    if isinstance(source, (str, Path)):
        source = get_local_source(id, source, **kwargs)

    elif isinstance(source, (list, pd.DataFrame)):
        source = MemoryDataset(id=id, vals=source)

    else:
        raise ValueError(f"Invalid source type: {type(source)}, We only support str, Path, list[dict], and pd.DataFrame")

    return source



from __future__ import annotations

import sys
from typing import Any, TypeAliasType

import pandas as pd
from pydantic import BaseModel, Field, create_model
from pydantic.fields import FieldInfo

from palimpzest.utils.hash_helpers import hash_for_serialized_dict

# DEFINITIONS
PANDAS_DTYPE_TO_PYDANTIC = {
    "object": str,
    "bool": bool,
    "int64": int,
    "float64": float,
}

# IMAGE TYPES
ImageFilepath = TypeAliasType('ImageFilepath', str)
ImageBase64 = TypeAliasType('ImageBase64', str)
ImageURL = TypeAliasType('ImageURL', str)

# AUDIO TYPES
AudioFilepath = TypeAliasType('AudioFilepath', str)
AudioBase64 = TypeAliasType('AudioBase64', str)

IMAGE_LIST_FIELD_TYPES = [
    list[ImageBase64],
    list[ImageFilepath],
    list[ImageURL],
    list[ImageBase64] | None,
    list[ImageFilepath] | None,
    list[ImageURL] | None,
    list[ImageBase64] | Any,
    list[ImageFilepath] | Any,
    list[ImageURL] | Any,
]
IMAGE_FIELD_TYPES = IMAGE_LIST_FIELD_TYPES + [
    ImageBase64, ImageFilepath, ImageURL,
    ImageBase64 | None, ImageFilepath | None, ImageURL | None,
    ImageBase64 | Any, ImageFilepath | Any, ImageURL | Any,
]
AUDIO_LIST_FIELD_TYPES = [
    list[AudioBase64],
    list[AudioFilepath],
    list[AudioBase64] | None,
    list[AudioFilepath] | None,
    list[AudioBase64] | Any,
    list[AudioFilepath] | Any,
]
AUDIO_FIELD_TYPES = AUDIO_LIST_FIELD_TYPES + [
    AudioBase64, AudioFilepath,
    AudioBase64 | None, AudioFilepath | None,
    AudioBase64 | Any, AudioFilepath | Any,
]


def get_schema_field_names(schema: type[BaseModel], id: str | None = None) -> list[str]:
    """Return the field names of a Pydantic model."""
    return list(schema.model_fields) if id is None else [f"{schema.__name__}.{id}.{field_name}" for field_name in schema.model_fields]


def _create_pickleable_model(fields: dict[str, tuple[type, FieldInfo]]) -> type[BaseModel]:
    """Create a Pydantic model that can be pickled."""
    # create unique name for the unioned model
    new_schema_name = f"Schema{sorted(fields.keys())}"
    new_schema_id = hash_for_serialized_dict({
        field_name: {"annotation": str(annotation), "default": str(field.default), "description": field.description}
        for field_name, (annotation, field) in fields.items()
    })

    # if this class already exists, get it from the module and return
    module = sys.modules[__name__]
    if hasattr(module, new_schema_id):
        return getattr(module, new_schema_id)

    # create the class dynamically
    new_model = create_model(new_schema_name, **fields)

    # register it in the module's namespace so pickle can find it
    module = sys.modules[__name__]
    setattr(module, new_schema_id, new_model)
    new_model.__module__ = module.__name__

    return new_model


def relax_schema(model: type[BaseModel]) -> type[BaseModel]:
    """Updates the type annotation for every field in the BaseModel to include typing.Any"""
    fields = {}
    for field_name, field in model.model_fields.items():
        fields[field_name] = (field.annotation | Any, field)

    return _create_pickleable_model(fields)


def project(model: type[BaseModel], project_fields: list[str]) -> type[BaseModel]:
    """Project a Pydantic model to only the specified columns."""
    # make sure projection column names are shortened
    project_fields = [field_name.split(".")[-1] for field_name in project_fields]

    # build up the fields for the new schema
    fields = {}
    for field_name, field in model.model_fields.items():
        if field_name in project_fields:
            fields[field_name] = (field.annotation, field)

    # create and return the new schema
    return _create_pickleable_model(fields)


def create_schema_from_fields(fields: list[dict]) -> type[BaseModel]:
    """Create a Pydantic model from a list of fields."""
    fields_ = {}
    for field in fields:
        assert "name" in field, "fields must contain a 'name' key"
        assert "type" in field, "fields must contain a 'type' key"
        assert "desc" in field or "description" in field, "fields must contain a 'description' key"

        # for backwards compatability, rename "desc" to "description"
        if "desc" in field:
            field["description"] = field.pop("desc")
        field_name = field["name"]
        field_type = field["type"]
        fields_[field_name] = (field_type, Field(**{k: v for k, v in field.items() if k not in ["name", "type"]}))

    return _create_pickleable_model(fields_)


def create_schema_from_df(df: pd.DataFrame) -> type[BaseModel]:
    """Create a Pydantic model from a Pandas DataFrame."""
    fields = {}
    for column, dtype in zip(df.columns, df.dtypes):
        column = f"column_{column}" if isinstance(column, int) else column
        field_desc = f"The {column} column from an input DataFrame"
        annotation = PANDAS_DTYPE_TO_PYDANTIC.get(str(dtype), Any)
        fields[column] = (annotation, Field(description=field_desc))

    # create and return the new schema
    return _create_pickleable_model(fields)


def union_schemas(models: list[type[BaseModel]], join: bool = False, on: list[str] | None = None) -> type[BaseModel]:
    """Union multiple Pydantic models into a single model."""
    # convert on to empty list if None
    if on is None:
        on = []

    # build up the fields for the new schema
    fields = {}
    for model in models:
        for field_name, field in model.model_fields.items():
            # for non-join unions, make sure duplicate fields have the same type
            if not join and field_name in fields:
                assert fields[field_name][0] == field.annotation, f"Field {field_name} has different types in different models"

            # for joins with "on" specified, no need to rename fields in "on"
            elif join and field_name in on and field_name in fields:
                continue

            # otherwise, rename duplicate fields by appending _right
            elif join and field_name in fields:
                while field_name in fields:
                    field_name = f"{field_name}_right"

            # add the field to the new schema
            fields[field_name] = (field.annotation, field)

    # create and return the new schema
    return _create_pickleable_model(fields)

###################################################################################
# "Core" useful Schemas. These are Schemas that almost everyone will need.
# File, TextFile, Image, PDF, etc.
###################################################################################


# First-level Schema's
class DefaultSchema(BaseModel):
    """Store context data."""
    value: Any = Field(description="The value of the input data")

class Download(BaseModel):
    """A download is a URL and the contents of the download."""
    url: str = Field(description="The URL of the download")
    content: bytes = Field(description="The contents of the download")
    timestamp: str = Field(description="The timestamp of the download")

class File(BaseModel):
    """
    A File is defined by two Fields:
    - the filename (string)
    - the contents of the file (bytes)
    """
    filename: str = Field(description="The UNIX-style name of the file")
    contents: bytes = Field(description="The contents of the file")

class TextFile(BaseModel):
    """A text file is a File that contains only text. No binary data."""
    filename: str = Field(description="The UNIX-style name of the file")
    contents: str = Field(description="The contents of the file")

class Average(BaseModel):
    average: float = Field(description="The average value of items in the dataset")

class Count(BaseModel):
    count: int = Field(description="The count of items in the dataset")

class Sum(BaseModel):
    sum: int = Field(description="The summation of items in the dataset")

class Min(BaseModel):
    min: int | float = Field(description="The minimum value of some items in the dataset")

class Max(BaseModel):
    max: int | float = Field(description="The maximum value of some items in the dataset")

class OperatorDerivedSchema(BaseModel):
    """Schema defined by an operator, e.g., a join or a group by"""

class Table(BaseModel):
    """A Table is an object composed of a header and rows."""
    filename: str = Field(description="The name of the file the table was extracted from")
    name: str = Field(description="The name of the table")
    header: list[str] = Field(description="The header of the table")
    rows: list[list] = Field(description="The rows of the table")

class URL(BaseModel):
    """A URL is a string that represents a web address."""
    url: str = Field(description="A URL")

class WebPage(BaseModel):
    """A web page is a URL and the contents of the page."""
    text: str = Field(description="The text contents of the web page")
    html: str = Field(description="The html contents of the web page")
    timestamp: str = Field(description="The timestamp of the download")
    filename: str = Field(description="The name of the file the web page was downloaded from")

# Second-level Schemas
class ImageFile(File):
    """A file that contains an image."""
    contents: ImageBase64 = Field(description="The contents of the image encoded as a base64 string")

class AudioFile(File):
    """A file that contains audio."""
    contents: AudioBase64 = Field(description="The contents of an audio recording encoded as a base64 string")

class PDFFile(File):
    """A PDF file is a File that is a PDF. It has specialized fields, font information, etc."""
    # This class is currently very impoverished. It needs a lot more fields before it can correctly represent a PDF.
    text_contents: str = Field(description="The text-only contents of the PDF")

class XLSFile(File):
    """An XLS file is a File that contains one or more Excel spreadsheets."""
    number_sheets: int = Field(description="The number of sheets in the Excel file")
    sheet_names: list[str] = Field(description="The names of the sheets in the Excel file")

# Third-level Schemas
class EquationImage(ImageFile):
    """An image that contains a mathematical equation."""
    equation_text: str = Field(description="The text representation of the equation in the image")

class PlotImage(ImageFile):
    """An image that contains a plot, such as a graph or chart."""
    plot_description: str = Field(description="A description of the plot")



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

from palimpzest.constants import Model
from palimpzest.policy import MaxQuality, Policy


# TODO: Add description for each field.
class QueryProcessorConfig(BaseModel):
    """Shared context for query processors"""
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # execution and optimization flags
    execution_strategy: str = Field(default="parallel")              # substituted with ExecutionStrategyType
    sentinel_execution_strategy: str | None = Field(default="auto")  # substituted with SentinelExecutionStrategyType
    optimizer_strategy: str = Field(default="pareto")                # substituted with OptimizationStrategyType

    # general execution flags
    policy: Policy = Field(default_factory=MaxQuality)
    enforce_types: bool = Field(default=False)
    scan_start_idx: int = Field(default=0)
    num_samples: int | None = Field(default=None)
    verbose: bool = Field(default=False)
    progress: bool = Field(default=True)
    available_models: list[Model] | None = Field(default=None)
    remove_models: list[Model] | None = Field(default=None)
    max_workers: int | None = Field(default=64)
    join_parallelism: int = Field(default=64)
    batch_size: int | None = Field(default=None)
    reasoning_effort: str | None = Field(default="default")  # Gemini: "disable", "low", "medium", "high"
    use_vertex: bool = Field(default=False)  # Whether to use Vertex models for Gemini or Google models
    gemini_credentials_path: str | None = Field(default=None)  # Path to Gemini credentials file
    api_base: str | None = Field(default=None)  # API base URL for vLLM

    # operator flags
    allow_bonded_query: bool = Field(default=True)
    allow_model_selection: bool = Field(default=True)
    allow_rag_reduction: bool = Field(default=True)
    allow_mixtures: bool = Field(default=True)
    allow_critic: bool = Field(default=True)
    allow_split_merge: bool = Field(default=False)
    use_final_op_quality: bool = Field(default=False)

    # sentinel optimization flags
    k: int = Field(default=6)
    j: int = Field(default=4)
    sample_budget: int = Field(default=100)
    sample_cost_budget: float | None = Field(default=None)
    seed: int = Field(default=42)
    exp_name: str | None = Field(default=None)
    priors: dict | None = Field(default=None)
    dont_use_priors: bool = Field(default=False)

    def to_dict(self) -> dict:
        """Convert the config to a dict representation."""
        return self.model_dump()

    def copy(self) -> QueryProcessorConfig:
        """Create a copy of the config."""
        return QueryProcessorConfig(**self.to_dict())



import json
import time

import litellm

# from colorama import Fore, Style
from palimpzest.constants import MODEL_CARDS, Cardinality, Model, PromptStrategy
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.models import GenerationStats
from palimpzest.prompts import (
    FLAT_MAP_IMAGE_VALIDATOR_PROMPT,
    FLAT_MAP_VALIDATOR_PROMPT,
    MAP_IMAGE_VALIDATOR_PROMPT,
    MAP_VALIDATOR_PROMPT,
    RETRIEVE_VALIDATOR_PROMPT,
    PromptFactory,
)
from palimpzest.query.generators.generators import get_json_from_answer
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import LLMFilter
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.topk import TopKOp


class Validator:
    """
    The Validator is used during optimization to score the output of physical operator(s) and physical plan(s).

    TODO: support end-to-end labels; will likely require a different SentinelExecutionStrategy which
          executes the full input to produce an output, evaluates the output, and then updates
          intermediate operator(s) based on the evaluation.
    """
    def __init__(self, model: Model = Model.o4_MINI):
        self.model = model
        self.filter_cache = {}
        self.join_cache = {}

    def map_score_fn(self, fields: list[str], input_record: dict, output: dict) -> float | None:
        raise NotImplementedError("Validator.map_score_fn not implemented.")

    def flat_map_score_fn(self, fields: list[str], input_record: dict, output: list[dict]) -> float | None:
        raise NotImplementedError("Validator.flat_map_score_fn not implemented.")

    def filter_score_fn(self, filter_str: str, input_record: dict, output: bool) -> float | None:
        raise NotImplementedError("Validator.filter_score_fn not implemented.")

    def join_score_fn(self, condition: str, left_input_record: dict, right_input_record: dict, output: bool) -> float | None:
        raise NotImplementedError("Validator.join_score_fn not implemented.")

    def topk_score_fn(self, fields: list[str], input_record: dict, output: dict) -> float | None:
        raise NotImplementedError("Validator.map_score_fn not implemented.")

    def _get_gen_stats_from_completion(self, completion, start_time: float) -> GenerationStats:
        """
        Extract generation stats from the given completion response.
        """
        usage = completion.usage.model_dump()

        # get cost per input/output token for the model and parse number of input and output tokens
        usd_per_input_token = MODEL_CARDS[self.model.value]["usd_per_input_token"]
        usd_per_output_token = MODEL_CARDS[self.model.value]["usd_per_output_token"]
        input_tokens = usage["prompt_tokens"]
        output_tokens = usage["completion_tokens"]

        return GenerationStats(
            model_name=self.model.value,
            llm_call_duration_secs=time.time() - start_time,
            fn_call_duration_secs=0.0,
            total_input_tokens=input_tokens,
            total_output_tokens=output_tokens,
            total_input_cost=input_tokens * usd_per_input_token,
            total_output_cost=output_tokens * usd_per_output_token,
            cost_per_record=input_tokens * usd_per_input_token + output_tokens * usd_per_output_token,
            total_llm_calls=1,
        )

    def _default_map_score_fn(self, op: LLMConvert, fields: list[str], input_record: DataRecord, output: dict) -> tuple[float | None, GenerationStats]:
        """
        Compute the quality of the generated output for the given fields and input_record.
        """
        # create prompt factory
        factory = PromptFactory(PromptStrategy.MAP, self.model, Cardinality.ONE_TO_ONE)

        # get the input messages; strip out the system message(s)
        msg_kwargs = {"output_schema": op.output_schema, "project_cols": op.get_input_fields()}
        messages = factory.create_messages(input_record, fields, **msg_kwargs)
        input_messages = [msg for msg in messages if msg["role"] != "system"]
        output = json.dumps(output, indent=2)
        output_message = f"OUTPUT:\n--------\n{output}\n\nEVALUATION: "
        # input_str = '\n'.join(list(map(lambda d: d['content'], input_messages + [{"role": "user", "content": output_message}])))

        # invoke the judge
        score, gen_stats = None, GenerationStats()
        try:
            start_time = time.time()
            validator_prompt = MAP_IMAGE_VALIDATOR_PROMPT if op.is_image_op() else MAP_VALIDATOR_PROMPT
            val_messages = [{"role": "system", "content": validator_prompt}] + input_messages + [{"role": "user", "content": output_message}]
            completion = litellm.completion(model=self.model.value, messages=val_messages)
            completion_text = completion.choices[0].message.content
            gen_stats = self._get_gen_stats_from_completion(completion, start_time)
            # print(f"INPUT:\n{input_str}")
            # print(Fore.GREEN + f"{completion_text}\n" + Style.RESET_ALL)

            # parse the evaluation
            eval_dict: dict = get_json_from_answer(completion_text, self.model, Cardinality.ONE_TO_ONE)
            score = sum(eval_dict.values()) / len(eval_dict)

        except Exception:
            pass

        return score, gen_stats

    def _default_flat_map_score_fn(self, op: LLMConvert, fields: list[str], input_record: dict, output: list[dict]) -> tuple[float | None, GenerationStats]:
        """
        Compute the quality for each record_op_stats object in the given record_set.
        """
        # create prompt factory
        factory = PromptFactory(PromptStrategy.MAP, self.model, Cardinality.ONE_TO_MANY)

        # get the input messages; strip out the system message(s)
        msg_kwargs = {"output_schema": op.output_schema, "project_cols": op.get_input_fields()}
        messages = factory.create_messages(input_record, fields, **msg_kwargs)
        input_messages = [msg for msg in messages if msg["role"] != "system"]
        output = json.dumps(output, indent=2)
        output_message = f"OUTPUTS:\n--------\n{output}\n\nEVALUATION: "
        # input_str = '\n'.join(list(map(lambda d: d['content'], input_messages + [{"role": "user", "content": output_message}])))

        # invoke the judge
        score, gen_stats = None, GenerationStats()
        try:
            start_time = time.time()
            validator_prompt = FLAT_MAP_IMAGE_VALIDATOR_PROMPT if op.is_image_op() else FLAT_MAP_VALIDATOR_PROMPT
            val_messages = [{"role": "system", "content": validator_prompt}] + input_messages + [{"role": "user", "content": output_message}]
            completion = litellm.completion(model="openai/o4-mini", messages=val_messages)
            completion_text = completion.choices[0].message.content
            gen_stats = self._get_gen_stats_from_completion(completion, start_time)
            # print(f"INPUT:\n{input_str}")
            # print(Fore.GREEN + f"{completion_text}\n" + Style.RESET_ALL)

            # parse the evaluation
            eval_dicts: list[dict] = get_json_from_answer(completion_text, self.model, Cardinality.ONE_TO_MANY)
            all_qualities = []
            for record_eval_dict in eval_dicts:
                all_qualities.extend(record_eval_dict.values())
            score = sum(all_qualities) / len(all_qualities)

        except Exception:
            pass

        return score, gen_stats

    def _default_filter_score_fn(self, op: LLMFilter, filter_str: str, input_record: dict, output: bool) -> tuple[float | None, GenerationStats]:
        """
        Compute the quality for each record_op_stats object in the given record_set.
        """
        score, gen_stats = None, GenerationStats()
        filter_input_hash = hash(f"{filter_str}{hash(input_record)}")
        label = self.filter_cache.get(filter_input_hash, None)
        if label is None:
            validator_op: LLMFilter = op.copy()
            validator_op.model = self.model
            try:
                target_record_set = validator_op(input_record)
                label = target_record_set[0]._passed_operator
                self.filter_cache[filter_input_hash] = label
                score = float(label == output)
                record_op_stats = target_record_set.record_op_stats[0]
                gen_stats = GenerationStats(
                    model_name=self.model.value,
                    total_input_tokens=record_op_stats.total_input_tokens,
                    total_output_tokens=record_op_stats.total_output_tokens,
                    total_input_cost=record_op_stats.total_input_cost,
                    total_output_cost=record_op_stats.total_output_cost,
                    cost_per_record=record_op_stats.cost_per_record,
                    llm_call_duration_secs=record_op_stats.llm_call_duration_secs,
                    fn_call_duration_secs=record_op_stats.fn_call_duration_secs,
                    total_llm_calls=record_op_stats.total_llm_calls,
                )

            except Exception:
                pass

        else:
            score = float(label == output)

        return score, gen_stats

    def _default_join_score_fn(self, op: JoinOp, condition: str, left_input_record: DataRecord, right_input_record: DataRecord, output: bool) -> tuple[float | None, GenerationStats]:
        score, gen_stats = None, GenerationStats()
        join_input_hash = hash(f"{condition}{hash(left_input_record)}{hash(right_input_record)}")
        label = self.join_cache.get(join_input_hash, None)
        if label is None:
            validator_op: JoinOp = op.copy()
            validator_op.model = self.model
            try:
                target_record_set, _ = validator_op([left_input_record], [right_input_record])
                label = target_record_set[0]._passed_operator
                self.join_cache[join_input_hash] = label
                score = float(label == output)
                record_op_stats = target_record_set.record_op_stats[0]
                gen_stats = GenerationStats(
                    model_name=self.model.value,
                    total_input_tokens=record_op_stats.total_input_tokens,
                    total_output_tokens=record_op_stats.total_output_tokens,
                    total_input_cost=record_op_stats.total_input_cost,
                    total_output_cost=record_op_stats.total_output_cost,
                    cost_per_record=record_op_stats.cost_per_record,
                    llm_call_duration_secs=record_op_stats.llm_call_duration_secs,
                    fn_call_duration_secs=record_op_stats.fn_call_duration_secs,
                    total_llm_calls=record_op_stats.total_llm_calls,
                )

            except Exception:
                pass

        else:
            score = float(label == output)

        return score, gen_stats

    def _default_topk_score_fn(self, op: TopKOp, fields: list[str], input_record: DataRecord, output: dict) -> tuple[float | None, GenerationStats]:
        """
        Compute the quality of the generated output for the given fields and input_record.
        """
        # TODO: top-k k=25; score each item based on relevance; compute F1
        # TODO: support retrieval over images
        # create prompt factory
        factory = PromptFactory(PromptStrategy.MAP, self.model, Cardinality.ONE_TO_ONE)

        # get the input messages; strip out the system message(s)
        msg_kwargs = {"output_schema": op.output_schema, "project_cols": op.get_input_fields()}
        messages = factory.create_messages(input_record, fields, **msg_kwargs)
        input_messages = [msg for msg in messages if msg["role"] != "system"]
        output = json.dumps(output, indent=2)
        output_message = f"OUTPUT:\n--------\n{output}\n\nEVALUATION: "
        # input_str = '\n'.join(list(map(lambda d: d['content'], input_messages + [{"role": "user", "content": output_message}])))

        # invoke the judge
        score, gen_stats = None, GenerationStats()
        try:
            start_time = time.time()
            # TODO: support retrieval over images
            validator_prompt = RETRIEVE_VALIDATOR_PROMPT
            val_messages = [{"role": "system", "content": validator_prompt}] + input_messages + [{"role": "user", "content": output_message}]
            completion = litellm.completion(model="openai/o4-mini", messages=val_messages)
            completion_text = completion.choices[0].message.content
            gen_stats = self._get_gen_stats_from_completion(completion, start_time)
            # print(f"INPUT:\n{input_str}")
            # print(Fore.GREEN + f"{completion_text}\n" + Style.RESET_ALL)

            # parse the evaluation
            eval_dict: dict = get_json_from_answer(completion_text, self.model, Cardinality.ONE_TO_ONE)
            score = sum(eval_dict.values()) / len(eval_dict)

        except Exception:
            pass

        return score, gen_stats


    def _score_map(self, op: LLMConvert, fields: list[str], input_record: DataRecord, output: dict, full_hash: str) -> tuple[float | None, GenerationStats, str]:
        try:
            out = self.map_score_fn(fields, input_record.to_dict(), output)
            score, gen_stats = out if isinstance(out, tuple) else (out, GenerationStats())
            return score, gen_stats, full_hash
        except NotImplementedError:
            score, gen_stats = self._default_map_score_fn(op, fields, input_record, output)
            return score, gen_stats, full_hash

    def _score_flat_map(self, op: LLMConvert, fields: list[str], input_record: DataRecord, output: list[dict], full_hash: str) -> tuple[float | None, GenerationStats, str]:
        try:
            out = self.flat_map_score_fn(fields, input_record.to_dict(), output)
            score, gen_stats = out if isinstance(out, tuple) else (out, GenerationStats())
            return score, gen_stats, full_hash
        except NotImplementedError:
            score, gen_stats = self._default_flat_map_score_fn(op, fields, input_record, output)
            return score, gen_stats, full_hash

    def _score_filter(self, op: LLMFilter, filter_str: str, input_record: DataRecord, output: bool, full_hash: str) -> tuple[float | None, GenerationStats, str]:
        try:
            out = self.filter_score_fn(filter_str, input_record.to_dict(), output)
            score, gen_stats = out if isinstance(out, tuple) else (out, GenerationStats())
            return score, gen_stats, full_hash
        except NotImplementedError:
            score, gen_stats = self._default_filter_score_fn(op, filter_str, input_record, output)
            return score, gen_stats, full_hash

    def _score_join(self, op: JoinOp, condition: str, left_input_record: DataRecord, right_input_record: DataRecord, output: bool, full_hash: str) -> tuple[float | None, GenerationStats, str]:
        try:
            out = self.join_score_fn(condition, left_input_record.to_dict(), right_input_record.to_dict(), output)
            score, gen_stats = out if isinstance(out, tuple) else (out, GenerationStats())
            return score, gen_stats, full_hash
        except NotImplementedError:
            score, gen_stats = self._default_join_score_fn(op, condition, left_input_record, right_input_record, output)
            return score, gen_stats, full_hash

    def _score_topk(self, op: TopKOp, fields: list[str], input_record: DataRecord, output: dict, full_hash: str) -> tuple[float | None, GenerationStats, str]:
        try:
            out = self.topk_score_fn(fields, input_record.to_dict(), output)
            score, gen_stats = out if isinstance(out, tuple) else (out, GenerationStats())
            return score, gen_stats, full_hash
        except NotImplementedError:
            score, gen_stats = self._default_topk_score_fn(op, fields, input_record, output)
            return score, gen_stats, full_hash



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Any

from palimpzest.constants import LOCAL_SCAN_TIME_PER_KB
from palimpzest.core.data import context
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import OperatorCostEstimates, RecordOpStats
from palimpzest.query.operators.physical import PhysicalOperator


class ScanPhysicalOp(PhysicalOperator, ABC):
    """
    Physical operators which implement root Datasets require slightly more information
    in order to accurately compute naive cost estimates. Thus, we use a slightly
    modified abstract base class for these operators.
    """
    # datasource: IterDataset
    def __init__(self, datasource: Any, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.datasource = datasource

    def __str__(self):
        op = f"{self.op_name()}({self.datasource}) -> {self.output_schema}\n"
        op += f"    ({', '.join(list(self.output_schema.model_fields))[:30]})\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"datasource_id": self.datasource.id, **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"datasource": self.datasource, **op_params}

    @abstractmethod
    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
        input_record_size_in_bytes: int | float,
    ) -> OperatorCostEstimates:
        """
        This function returns a naive estimate of this operator's:
        - cardinality
        - time_per_record
        - cost_per_record
        - quality

        For the implemented operator. These will be used by the CostModel
        when PZ does not have sample execution data -- and it will be necessary
        in some cases even when sample execution data is present. (For example,
        the cardinality of each operator cannot be estimated based on sample
        execution data alone -- thus ScanPhysicalOps need to give
        at least ballpark correct estimates of this quantity).
        """
        pass

    def __call__(self, idx: int) -> DataRecordSet:
        """
        This function invokes `self.datasource.__getitem__` on the given `idx` to retrieve the next data item.
        It then returns this item as a DataRecord wrapped in a DataRecordSet.
        """
        start_time = time.time()
        item = self.datasource[idx]
        end_time = time.time()

        # check that item covers fields in output schema
        output_field_names = list(self.output_schema.model_fields)
        assert all([field in item for field in output_field_names]), f"Some fields in Dataset schema not present in item!\n - Dataset fields: {output_field_names}\n - Item fields: {list(item.keys())}"

        # construct a DataRecord from the item
        data_item = self.output_schema(**{field: item[field] for field in output_field_names})
        dr = DataRecord(data_item, source_indices=[f"{self.datasource.id}-{idx}"])

        # create RecordOpStats objects
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=(end_time - start_time),
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )
 
        # construct and return DataRecordSet object
        return DataRecordSet([dr], [record_op_stats])


class MarshalAndScanDataOp(ScanPhysicalOp):
    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
        input_record_size_in_bytes: int | float,
    ) -> OperatorCostEstimates:
        # get inputs needed for naive cost estimation
        # TODO: we should rename cardinality --> "multiplier" or "selectivity" one-to-one / one-to-many

        # estimate time spent reading each record
        per_record_size_kb = input_record_size_in_bytes / 1024.0

        # TODO: cannot do the first computation b/c we cannot import iter_dataset; possibly revisit
        # time_per_record = (
        #     MEMORY_SCAN_TIME_PER_KB * per_record_size_kb
        #     if isinstance(self.datasource, (iter_dataset.MemoryDataset))
        #     else LOCAL_SCAN_TIME_PER_KB * per_record_size_kb
        # )
        time_per_record = LOCAL_SCAN_TIME_PER_KB * per_record_size_kb

        # estimate output cardinality
        cardinality = source_op_cost_estimates.cardinality

        # for now, assume no cost per record for reading data
        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0,
            quality=1.0,
        )


class ContextScanOp(PhysicalOperator):
    """
    Physical operator which facillitates the loading of a Context for processing.
    """

    def __init__(self, context: context.Context, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = context

    def __str__(self):
        op = f"{self.op_name()}({self.context}) -> {self.output_schema}\n"
        op += f"    ({', '.join(list(self.output_schema.model_fields))[:30]})\n"
        return op

    def get_id_params(self):
        return super().get_id_params()

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"context": self.context, **op_params}

    def naive_cost_estimates(
        self,
        source_op_cost_estimates: OperatorCostEstimates,
    ):
        # get inputs needed for naive cost estimation
        # TODO: we should rename cardinality --> "multiplier" or "selectivity" one-to-one / one-to-many

        # estimate time spent reading each record
        time_per_record = LOCAL_SCAN_TIME_PER_KB * 1.0

        # for now, assume no cost per record for reading data
        return OperatorCostEstimates(
            cardinality=1.0,
            time_per_record=time_per_record,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, *args, **kwargs) -> DataRecordSet:
        """
        This function returns the context as a DataRecord wrapped in a DataRecordSet.
        """
        # construct a DataRecord from the context
        start_time = time.time()
        dr = DataRecord(self.output_schema(), source_indices=[f"{self.context.id}-{0}"])
        dr.context = self.context
        end_time = time.time()

        # create RecordOpStats objects
        record_op_stats = RecordOpStats(
            record_id=dr._id,
            record_parent_ids=dr._parent_ids,
            record_source_indices=dr._source_indices,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=(end_time - start_time),
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )
 
        # construct and return DataRecordSet object
        return DataRecordSet([dr], [record_op_stats])



"""This file contains factory methods which return template prompts and return messages for chat payloads."""

import base64
import json
from typing import Any

from pydantic import BaseModel

from palimpzest.constants import (
    LLAMA_CONTEXT_TOKENS_LIMIT,
    TOKENS_PER_CHARACTER,
    Cardinality,
    Modality,
    Model,
    PromptStrategy,
)
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.schemas import (
    AUDIO_FIELD_TYPES,
    IMAGE_FIELD_TYPES,
    AudioBase64,
    AudioFilepath,
    ImageBase64,
    ImageFilepath,
    ImageURL,
)
from palimpzest.prompts.aggregate_prompts import (
    AGG_BASE_SYSTEM_PROMPT,
    AGG_BASE_USER_PROMPT,
    AGG_NO_REASONING_BASE_SYSTEM_PROMPT,
    AGG_NO_REASONING_BASE_USER_PROMPT,
)
from palimpzest.prompts.convert_prompts import (
    MAP_BASE_SYSTEM_PROMPT,
    MAP_BASE_USER_PROMPT,
    MAP_NO_REASONING_BASE_SYSTEM_PROMPT,
    MAP_NO_REASONING_BASE_USER_PROMPT,
)
from palimpzest.prompts.critique_and_refine_prompts import (
    BASE_CRITIQUE_PROMPT,
    BASE_REFINEMENT_PROMPT,
    FILTER_CRITIQUE_CRITERIA,
    FILTER_CRITIQUE_FINISH_INSTRUCTION,
    FILTER_REFINEMENT_CRITERIA,
    FILTER_REFINEMENT_FINISH_INSTRUCTION,
    MAP_CRITIQUE_CRITERIA,
    MAP_CRITIQUE_FINISH_INSTRUCTION,
    MAP_REFINEMENT_CRITERIA,
    MAP_REFINEMENT_FINISH_INSTRUCTION,
)
from palimpzest.prompts.filter_prompts import (
    FILTER_BASE_SYSTEM_PROMPT,
    FILTER_BASE_USER_PROMPT,
    FILTER_NO_REASONING_BASE_SYSTEM_PROMPT,
    FILTER_NO_REASONING_BASE_USER_PROMPT,
)
from palimpzest.prompts.join_prompts import (
    JOIN_BASE_SYSTEM_PROMPT,
    JOIN_BASE_USER_PROMPT,
    JOIN_NO_REASONING_BASE_SYSTEM_PROMPT,
    JOIN_NO_REASONING_BASE_USER_PROMPT,
)
from palimpzest.prompts.moa_aggregator_prompts import (
    FILTER_MOA_AGG_BASE_SYSTEM_PROMPT,
    FILTER_MOA_AGG_BASE_USER_PROMPT,
    MAP_MOA_AGG_BASE_SYSTEM_PROMPT,
    MAP_MOA_AGG_BASE_USER_PROMPT,
)
from palimpzest.prompts.moa_proposer_prompts import (
    FILTER_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
    FILTER_MOA_PROPOSER_BASE_USER_PROMPT,
    MAP_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
    MAP_MOA_PROPOSER_BASE_USER_PROMPT,
)
from palimpzest.prompts.split_merge_prompts import (
    FILTER_SPLIT_MERGER_BASE_SYSTEM_PROMPT,
    FILTER_SPLIT_MERGER_BASE_USER_PROMPT,
    MAP_SPLIT_MERGER_BASE_SYSTEM_PROMPT,
    MAP_SPLIT_MERGER_BASE_USER_PROMPT,
)
from palimpzest.prompts.split_proposer_prompts import (
    FILTER_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
    FILTER_SPLIT_PROPOSER_BASE_USER_PROMPT,
    MAP_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
    MAP_SPLIT_PROPOSER_BASE_USER_PROMPT,
)
from palimpzest.prompts.utils import (
    AGG_AUDIO_DISCLAIMER,
    AGG_EXAMPLE_ANSWER,
    AGG_EXAMPLE_OUTPUT_FIELDS,
    AGG_EXAMPLE_REASONING,
    AGG_IMAGE_DISCLAIMER,
    AGG_JOB_INSTRUCTION,
    AUDIO_DISCLAIMER,
    AUDIO_EXAMPLE_ANSWER,
    AUDIO_EXAMPLE_CONTEXT,
    AUDIO_EXAMPLE_INPUT_FIELDS,
    AUDIO_EXAMPLE_OUTPUT_FIELDS,
    AUDIO_EXAMPLE_REASONING,
    AUDIO_SENTENCE_EXAMPLE_ANSWER,
    DESC_SECTION,
    EXAMPLE_AGG_INSTRUCTION,
    EXAMPLE_FILTER_CONDITION,
    EXAMPLE_JOIN_CONDITION,
    FILTER_EXAMPLE_REASONING,
    FILTER_JOB_INSTRUCTION,
    IMAGE_DISCLAIMER,
    IMAGE_EXAMPLE_ANSWER,
    IMAGE_EXAMPLE_CONTEXT,
    IMAGE_EXAMPLE_INPUT_FIELDS,
    IMAGE_EXAMPLE_OUTPUT_FIELDS,
    IMAGE_EXAMPLE_REASONING,
    IMAGE_SENTENCE_EXAMPLE_ANSWER,
    JOIN_EXAMPLE_REASONING,
    JOIN_JOB_INSTRUCTION,
    MAP_JOB_INSTRUCTION,
    ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION,
    ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION,
    PROPOSER_JOB_INSTRUCTION,
    RIGHT_AUDIO_DISCLAIMER,
    RIGHT_AUDIO_EXAMPLE_CONTEXT,
    RIGHT_AUDIO_EXAMPLE_INPUT_FIELDS,
    RIGHT_IMAGE_DISCLAIMER,
    RIGHT_IMAGE_EXAMPLE_CONTEXT,
    RIGHT_IMAGE_EXAMPLE_INPUT_FIELDS,
    RIGHT_TEXT_EXAMPLE_CONTEXT,
    RIGHT_TEXT_EXAMPLE_INPUT_FIELDS,
    SECOND_AUDIO_EXAMPLE_CONTEXT,
    SECOND_IMAGE_EXAMPLE_CONTEXT,
    SECOND_TEXT_EXAMPLE_CONTEXT,
    TEXT_EXAMPLE_ANSWER,
    TEXT_EXAMPLE_CONTEXT,
    TEXT_EXAMPLE_INPUT_FIELDS,
    TEXT_EXAMPLE_OUTPUT_FIELDS,
    TEXT_EXAMPLE_REASONING,
    TEXT_SENTENCE_EXAMPLE_ANSWER,
    THIRD_AUDIO_EXAMPLE_CONTEXT,
    THIRD_IMAGE_EXAMPLE_CONTEXT,
    THIRD_TEXT_EXAMPLE_CONTEXT,
)


class PromptFactory:
    """Factory class for generating prompts for the Generator given the input(s)."""

    BASE_SYSTEM_PROMPT_MAP = {
        # agg user prompts
        PromptStrategy.AGG: AGG_BASE_SYSTEM_PROMPT,
        PromptStrategy.AGG_NO_REASONING: AGG_NO_REASONING_BASE_SYSTEM_PROMPT,

        # filter system prompts
        PromptStrategy.FILTER: FILTER_BASE_SYSTEM_PROMPT,
        PromptStrategy.FILTER_NO_REASONING: FILTER_NO_REASONING_BASE_SYSTEM_PROMPT,
        PromptStrategy.FILTER_CRITIC: None,
        PromptStrategy.FILTER_REFINE: None,
        PromptStrategy.FILTER_MOA_PROPOSER: FILTER_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.FILTER_MOA_AGG: FILTER_MOA_AGG_BASE_SYSTEM_PROMPT,
        PromptStrategy.FILTER_SPLIT_PROPOSER: FILTER_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.FILTER_SPLIT_MERGER: FILTER_SPLIT_MERGER_BASE_SYSTEM_PROMPT,

        # join system prompts
        PromptStrategy.JOIN: JOIN_BASE_SYSTEM_PROMPT,
        PromptStrategy.JOIN_NO_REASONING: JOIN_NO_REASONING_BASE_SYSTEM_PROMPT,

        # map system prompts
        PromptStrategy.MAP: MAP_BASE_SYSTEM_PROMPT,
        PromptStrategy.MAP_NO_REASONING: MAP_NO_REASONING_BASE_SYSTEM_PROMPT,
        PromptStrategy.MAP_CRITIC: None,
        PromptStrategy.MAP_REFINE: None,
        PromptStrategy.MAP_MOA_PROPOSER: MAP_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.MAP_MOA_AGG: MAP_MOA_AGG_BASE_SYSTEM_PROMPT,
        PromptStrategy.MAP_SPLIT_PROPOSER: MAP_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.MAP_SPLIT_MERGER: MAP_SPLIT_MERGER_BASE_SYSTEM_PROMPT,
    }
    BASE_USER_PROMPT_MAP = {
        # agg user prompts
        PromptStrategy.AGG: AGG_BASE_USER_PROMPT,
        PromptStrategy.AGG_NO_REASONING: AGG_NO_REASONING_BASE_USER_PROMPT,

        # filter user prompts
        PromptStrategy.FILTER: FILTER_BASE_USER_PROMPT,
        PromptStrategy.FILTER_NO_REASONING: FILTER_NO_REASONING_BASE_USER_PROMPT,
        PromptStrategy.FILTER_CRITIC: BASE_CRITIQUE_PROMPT,
        PromptStrategy.FILTER_REFINE: BASE_REFINEMENT_PROMPT,
        PromptStrategy.FILTER_MOA_PROPOSER: FILTER_MOA_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.FILTER_MOA_AGG: FILTER_MOA_AGG_BASE_USER_PROMPT,
        PromptStrategy.FILTER_SPLIT_PROPOSER: FILTER_SPLIT_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.FILTER_SPLIT_MERGER: FILTER_SPLIT_MERGER_BASE_USER_PROMPT,

        # join user prompts
        PromptStrategy.JOIN: JOIN_BASE_USER_PROMPT,
        PromptStrategy.JOIN_NO_REASONING: JOIN_NO_REASONING_BASE_USER_PROMPT,

        # map user prompts
        PromptStrategy.MAP: MAP_BASE_USER_PROMPT,
        PromptStrategy.MAP_NO_REASONING: MAP_NO_REASONING_BASE_USER_PROMPT,
        PromptStrategy.MAP_CRITIC: BASE_CRITIQUE_PROMPT,
        PromptStrategy.MAP_REFINE: BASE_REFINEMENT_PROMPT,
        PromptStrategy.MAP_MOA_PROPOSER: MAP_MOA_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.MAP_MOA_AGG: MAP_MOA_AGG_BASE_USER_PROMPT,
        PromptStrategy.MAP_SPLIT_PROPOSER: MAP_SPLIT_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.MAP_SPLIT_MERGER: MAP_SPLIT_MERGER_BASE_USER_PROMPT,
    }

    def __init__(self, prompt_strategy: PromptStrategy, model: Model, cardinality: Cardinality, desc: str | None = None) -> None:
        self.prompt_strategy = prompt_strategy
        self.model = model
        self.cardinality = cardinality
        self.desc = desc

    def _get_context(self, candidate: DataRecord | list[DataRecord], input_fields: list[str]) -> str:
        """
        Returns the context for the prompt.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.

        Returns:
            str: The context.
        """
        # TODO: remove mask_filepaths=True after SemBench evaluation
        # get context from input record (project_cols will be None if not provided in kwargs)
        if isinstance(candidate, list):
            context: list[dict] = [record.to_dict(include_bytes=False, project_cols=input_fields, mask_filepaths=True) for record in candidate]
        else:
            context: dict = candidate.to_dict(include_bytes=False, project_cols=input_fields, mask_filepaths=True)

        # TODO: MOVE THIS LOGIC INTO A CHUNKING / CONTEXT MANAGEMENT CLASS
        #   - this class should be able to:
        #      - handle the context length of different models (i.e. self.model should be an input)
        #      - handle images
        #      - handle the issue with `original_messages` (ask Matt if this is not clear)
        # TODO: this does not work for image prompts
        # TODO: this ignores the size of the `orignal_messages` in critique and refine prompts
        # NOTE: llama models are disallowed for aggregation so we can assume context is a dict here
        # cut down on context based on window length
        if self.model.is_llama_model():
            assert isinstance(context, dict), "Llama models are not allowed for aggregation operations."
            total_context_len = len(json.dumps(context, indent=2))

            # sort fields by length and progressively strip from the longest field until it is short enough;
            # NOTE: LLAMA_CONTEXT_TOKENS_LIMIT is a rough estimate which leaves room for the rest of the prompt text
            while total_context_len * TOKENS_PER_CHARACTER > LLAMA_CONTEXT_TOKENS_LIMIT:
                # sort fields by length
                field_lengths = [(field, len(value) if value is not None else 0) for field, value in context.items()]
                sorted_fields = sorted(field_lengths, key=lambda item: item[1], reverse=True)

                # get field with longest context
                longest_field_name, longest_field_length = sorted_fields[0]

                # trim the field
                context_factor = LLAMA_CONTEXT_TOKENS_LIMIT / (total_context_len * TOKENS_PER_CHARACTER)
                keep_frac_idx = int(longest_field_length * context_factor)
                context[longest_field_name] = context[longest_field_name][:keep_frac_idx]

                # update total context length
                total_context_len = len(json.dumps(context, indent=2))

        return json.dumps(context, indent=2)

    def _get_input_fields(self, candidate: DataRecord, **kwargs) -> list[str]:
        """
        The list of input fields to be templated into the prompt(s).
        If the user provides a list of "project_cols" in kwargs, then this list will be returned.
        Otherwise, this function returns the list of all field names in the candidate record.

        Args:
            candidate (DataRecord): The input record.
            kwargs: The keyword arguments provided by the user.

        Returns:
            list[str]: The list of input field names.
        """
        # NOTE: joins will include left and right input fields in project_cols, so we have to check
        #       if the field is in the candidate record
        input_fields = kwargs.get("project_cols", candidate.get_field_names())
        input_fields = [field for field in input_fields if field in candidate.get_field_names()]
        return input_fields

    def _get_input_modalities(self, candidate: DataRecord, input_fields: list[str]) -> set[Modality]:
        """
        The list of input modalities for the given input fields.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.

        Returns:
            set[Modality]: The list of input modalities.
        """
        input_modalities = []
        for field_name in input_fields:
            field_type = candidate.get_field_type(field_name)
            if field_type.annotation in IMAGE_FIELD_TYPES:
                input_modalities.append(Modality.IMAGE)
            elif field_type.annotation in AUDIO_FIELD_TYPES:
                input_modalities.append(Modality.AUDIO)
            else:
                input_modalities.append(Modality.TEXT)

        return set(input_modalities)

    def _get_modalities_str(self, input_modalities: set[Modality]) -> str:
        """
        Returns a format string to reflect the input modalities.

        Args:
            input_modalities (set[Modality]): The input modalities.

        Returns:
            str: The string to reflect the input modalities.
        """
        if input_modalities == {Modality.TEXT}:
            return "text"
        elif input_modalities == {Modality.IMAGE}:
            return "image(s)"
        elif input_modalities == {Modality.AUDIO}:
            return "audio"
        elif input_modalities == {Modality.TEXT, Modality.IMAGE}:
            return "text and/or image(s)"
        elif input_modalities == {Modality.TEXT, Modality.AUDIO}:
            return "text and/or audio"
        elif input_modalities == {Modality.IMAGE, Modality.AUDIO}:
            return "image(s) and/or audio"
        elif input_modalities == {Modality.TEXT, Modality.IMAGE, Modality.AUDIO}:
            return "text, image(s), and/or audio"

    def _get_input_fields_desc(self, candidate: DataRecord, input_fields: list[str]) -> str:
        """
        Returns a multi-line description of each input field for the prompt.

        Args:
            input_fields (list[str]): The input fields.

        Returns:
            str: The input fields description.
        """
        input_fields_desc = ""
        for field_name in input_fields:
            input_fields_desc += f"- {field_name}: {candidate.get_field_type(field_name).description}\n"

        return input_fields_desc[:-1]

    def _get_output_fields_desc(self, output_fields: list[str], **kwargs) -> str:
        """
        Returns a multi-line description of each output field for the prompt.

        Args:
            output_fields (list[str]): The output fields.
            kwargs: The keyword arguments provided by the user.

        Returns:
            str: The output fields description.
        """
        output_fields_desc = ""
        output_schema: type[BaseModel] = kwargs.get("output_schema")
        if self.prompt_strategy.is_map_prompt() or self.prompt_strategy.is_agg_prompt():
            assert output_schema is not None, "Output schema must be provided for convert prompts."

            for field_name in sorted(output_fields):
                desc = output_schema.model_fields[field_name].description
                output_fields_desc += f"- {field_name}: {'no description available' if desc is None else desc}\n"

        # strip the last newline characters from the field descriptions and return
        return output_fields_desc[:-1]

    def _get_agg_instruction(self, **kwargs) -> str | None:
        """
        Returns the aggregation instruction for the aggregation operation.

        Returns:
            str | None: The aggregation instruction (if applicable).
        """
        agg_instruction = kwargs.get("agg_instruction")
        if self.prompt_strategy.is_agg_prompt():
            assert agg_instruction is not None, "Aggregation instruction must be provided for aggregation operations."

        return agg_instruction

    def _get_filter_condition(self, **kwargs) -> str | None:
        """
        Returns the filter condition for the filter operation.

        Returns:
            str | None: The filter condition (if applicable).
        """
        filter_condition = kwargs.get("filter_condition")
        if self.prompt_strategy.is_filter_prompt():
            assert filter_condition is not None, "Filter condition must be provided for filter operations."

        return filter_condition

    def _get_join_condition(self, **kwargs) -> str | None:
        """
        Returns the join condition for the join operation.

        Returns:
            str | None: The join condition (if applicable).
        """
        join_condition = kwargs.get("join_condition")
        if self.prompt_strategy.is_join_prompt():
            assert join_condition is not None, "Join condition must be provided for join operations."

        return join_condition

    def _get_original_output(self, **kwargs) -> str | None:
        """
        Returns the original output from a previous model generation for the critique and refinement operations.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The original output.
        """
        original_output = kwargs.get("original_output")
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            assert original_output is not None, (
                "Original output must be provided for critique and refinement operations."
            )

        return original_output

    def _get_critique_output(self, **kwargs) -> str | None:
        """
        Returns the critique output for the refinement operation.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The critique output.
        """
        critique_output = kwargs.get("critique_output")
        if self.prompt_strategy.is_refine_prompt():
            assert critique_output is not None, "Critique output must be provided for refinement operations."

        return critique_output

    def _get_model_responses(self, **kwargs) -> str | None:
        """
        Returns the model responses for the mixture-of-agents aggregation operation.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The model responses.
        """
        model_responses = None
        if self.prompt_strategy.is_moa_aggregator_prompt():
            model_responses = ""
            for idx, model_response in enumerate(kwargs.get("model_responses")):
                model_responses += f"MODEL RESPONSE {idx + 1}: {model_response.rstrip()}\n\n"
        model_responses = model_responses.rstrip() if model_responses is not None else None

        return model_responses

    def _get_chunk_outputs(self, **kwargs) -> str | None:
        """
        Returns the chunk outputs for the split-convert.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The chunk outputs.
        """
        chunk_outputs = None
        if self.prompt_strategy.is_split_merger_prompt():
            chunk_outputs = ""
            for idx, chunk_output in enumerate(kwargs.get("chunk_outputs")):
                chunk_outputs += f"CHUNK OUTPUT {idx + 1}: {chunk_output.rstrip()}\n\n"
        chunk_outputs = chunk_outputs.rstrip() if chunk_outputs is not None else None

        return chunk_outputs

    def _get_output_format_instruction(self) -> str:
        """
        Returns the output format instruction based on the cardinality.

        Returns:
            str: The output format instruction.
        """
        return (
            ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION
            if self.cardinality == Cardinality.ONE_TO_ONE
            else ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION
        )

    def _get_job_instruction(self, input_modalities: set[Modality]) -> str | None:
        """
        Returns the job instruction based on the prompt strategy.

        Args:
            input_modalities (set[Modality]): The modalities of the input fields.

        Returns:
            str | None: The job instruction.
        """
        # get the job instruction based on the prompt strategy
        job_instruction = None
        if self.prompt_strategy.is_moa_proposer_prompt() or self.prompt_strategy.is_split_proposer_prompt():
            job_instruction = PROPOSER_JOB_INSTRUCTION
        elif self.prompt_strategy.is_map_prompt():
            job_instruction = MAP_JOB_INSTRUCTION
        elif self.prompt_strategy.is_filter_prompt():
            job_instruction = FILTER_JOB_INSTRUCTION
        elif self.prompt_strategy.is_join_prompt():
            job_instruction = JOIN_JOB_INSTRUCTION
        elif self.prompt_strategy.is_agg_prompt():
            job_instruction = AGG_JOB_INSTRUCTION

        # format the job instruction based on the input modalities
        modalities = self._get_modalities_str(input_modalities)
        if job_instruction is not None:
            job_instruction = job_instruction.format(modalities=modalities)

        return job_instruction

    def _get_desc_section(self) -> str:
        """
        Returns the description section for the prompt.

        Returns:
            str: The description section (if applicable).
        """
        desc_section = ""
        if self.desc is not None:
            desc_section = DESC_SECTION.format(desc=self.desc)

        return desc_section

    def _get_critique_criteria(self) -> str | None:
        """
        Returns the critique criteria for the critique operation.

        Returns:
            str | None: The critique criteria (if applicable).
        """
        critique_criteria = None
        if self.prompt_strategy.is_critic_prompt():
            critique_criteria = MAP_CRITIQUE_CRITERIA if self.prompt_strategy.is_map_prompt() else FILTER_CRITIQUE_CRITERIA

        return critique_criteria

    def _get_refinement_criteria(self) -> str | None:
        """
        Returns the refinement criteria for the refinement operation.

        Returns:
            str | None: The refinement criteria (if applicable).
        """
        refinement_criteria = None
        if self.prompt_strategy.is_refine_prompt():
            refinement_criteria = MAP_REFINEMENT_CRITERIA if self.prompt_strategy.is_map_prompt() else FILTER_REFINEMENT_CRITERIA

        return refinement_criteria

    def _get_finish_instruction(self) -> str | None:
        """
        Returns the finish instruction for the critique and refinement operations.

        Returns:
            str | None: The finish instruction (if applicable).
        """
        finish_instruction = None
        if self.prompt_strategy.is_critic_prompt():
            finish_instruction = MAP_CRITIQUE_FINISH_INSTRUCTION if self.prompt_strategy.is_map_prompt() else FILTER_CRITIQUE_FINISH_INSTRUCTION
        elif self.prompt_strategy.is_refine_prompt():
            finish_instruction = MAP_REFINEMENT_FINISH_INSTRUCTION if self.prompt_strategy.is_map_prompt() else FILTER_REFINEMENT_FINISH_INSTRUCTION

        return finish_instruction

    def _get_example_input_fields(self, input_modalities: set[Modality], right: bool = False) -> str:
        """
        Returns the example input fields for the prompt.

        Args:
            input_modalities (set[Modality]): The modalities of the input fields.
            right (bool): Whether to return the right input fields for the join prompt.

        Returns:
            str: The example input fields.
        """
        input_modality_to_example_input_fields = {
            Modality.TEXT: RIGHT_TEXT_EXAMPLE_INPUT_FIELDS if right else TEXT_EXAMPLE_INPUT_FIELDS,
            Modality.IMAGE: RIGHT_IMAGE_EXAMPLE_INPUT_FIELDS if right else IMAGE_EXAMPLE_INPUT_FIELDS,
            Modality.AUDIO: RIGHT_AUDIO_EXAMPLE_INPUT_FIELDS if right else AUDIO_EXAMPLE_INPUT_FIELDS,
        }

        example_input_fields = ""
        for input_modality in input_modalities:
            example_input_fields += input_modality_to_example_input_fields[input_modality].rstrip()
        example_input_fields = example_input_fields.lstrip() + "\n"

        return example_input_fields

    def _get_example_output_fields(self, input_modalities: set[Modality]) -> str:
        """
        Returns the example output fields for the prompt.

        Returns:
            str: The example output fields.
        """
        if self.prompt_strategy.is_agg_prompt():
            return AGG_EXAMPLE_OUTPUT_FIELDS

        input_modality_to_example_output_fields = {
            Modality.TEXT: TEXT_EXAMPLE_OUTPUT_FIELDS,
            Modality.IMAGE: IMAGE_EXAMPLE_OUTPUT_FIELDS,
            Modality.AUDIO: AUDIO_EXAMPLE_OUTPUT_FIELDS,
        }

        example_output_fields = ""
        for input_modality in input_modalities:
            example_output_fields += input_modality_to_example_output_fields[input_modality].rstrip()
        example_output_fields = example_output_fields.lstrip() + "\n"

        return example_output_fields

    def _get_example_context(self, input_modalities: set[Modality], right: bool = False, second: bool = False, third: bool = False) -> str:
        """
        Returns the example context for the prompt.

        Returns:
            str: The example context.
        """
        assert not (second and third), "Cannot have both second and third example contexts."
        assert not (right and (second or third)), "Right context is only used for joins; second and third contexts only use for aggregations."
        text_example_context = TEXT_EXAMPLE_CONTEXT
        image_example_context = IMAGE_EXAMPLE_CONTEXT
        audio_example_context = AUDIO_EXAMPLE_CONTEXT
        if second:
            text_example_context = SECOND_TEXT_EXAMPLE_CONTEXT
            image_example_context = SECOND_IMAGE_EXAMPLE_CONTEXT
            audio_example_context = SECOND_AUDIO_EXAMPLE_CONTEXT
        elif third:
            text_example_context = THIRD_TEXT_EXAMPLE_CONTEXT
            image_example_context = THIRD_IMAGE_EXAMPLE_CONTEXT
            audio_example_context = THIRD_AUDIO_EXAMPLE_CONTEXT

        input_modality_to_example_context = {
            Modality.TEXT: RIGHT_TEXT_EXAMPLE_CONTEXT if right else text_example_context,
            Modality.IMAGE: RIGHT_IMAGE_EXAMPLE_CONTEXT if right else image_example_context,
            Modality.AUDIO: RIGHT_AUDIO_EXAMPLE_CONTEXT if right else audio_example_context,
        }

        example_context = ""
        for input_modality in input_modalities:
            example_context += input_modality_to_example_context[input_modality].rstrip() + ","
        example_context = example_context[:-1] + "\n"

        return example_context

    def _get_image_disclaimer(self, input_modalities: set[Modality], right: bool = False, agg: bool = False) -> str:
        """
        Returns the image disclaimer for the prompt. The disclaimer must be an empty string
        for non-image prompts.

        Returns:
            str: The image disclaimer. If this is a text prompt then it is an empty string.
        """
        assert not (right and agg), "Right image disclaimer is only used for joins; agg image disclaimer only used for aggregations."
        image_disclaimer = AGG_IMAGE_DISCLAIMER if agg else IMAGE_DISCLAIMER
        image_disclaimer = RIGHT_IMAGE_DISCLAIMER if right else image_disclaimer
        return image_disclaimer if Modality.IMAGE in input_modalities else ""

    def _get_audio_disclaimer(self, input_modalities: set[Modality], right: bool = False, agg: bool = False) -> str:
        """
        Returns the audio disclaimer for the prompt. The disclaimer must be an empty string
        for non-audio prompts.

        Returns:
            str: The audio disclaimer. If this is a text prompt then it is an empty string.
        """
        assert not (right and agg), "Right audio disclaimer is only used for joins; agg audio disclaimer only used for aggregations."
        audio_disclaimer = AGG_AUDIO_DISCLAIMER if agg else AUDIO_DISCLAIMER
        audio_disclaimer = RIGHT_AUDIO_DISCLAIMER if right else audio_disclaimer
        return audio_disclaimer if Modality.AUDIO in input_modalities else ""

    def _get_example_reasoning(self, input_modalities: set[Modality]) -> str:
        """
        Returns the example reasoning for the prompt.

        Returns:
            str: The example reasoning.
        """
        if self.prompt_strategy.is_filter_prompt():
            return FILTER_EXAMPLE_REASONING
        elif self.prompt_strategy.is_join_prompt():
            return JOIN_EXAMPLE_REASONING
        elif self.prompt_strategy.is_agg_prompt():
            return AGG_EXAMPLE_REASONING

        input_modality_to_example_reasoning = {
            Modality.TEXT: TEXT_EXAMPLE_REASONING,
            Modality.IMAGE: IMAGE_EXAMPLE_REASONING,
            Modality.AUDIO: AUDIO_EXAMPLE_REASONING,
        }

        example_reasoning = ""
        for input_modality in input_modalities:
            example_reasoning += input_modality_to_example_reasoning[input_modality] + " "
        example_reasoning = example_reasoning.rstrip()

        return example_reasoning

    def _get_example_answer(self, input_modalities: set[Modality]) -> str:
        """
        Returns the example answer for the prompt.

        Returns:
            str: The example answer.
        """
        if self.prompt_strategy.is_agg_prompt():
            return AGG_EXAMPLE_ANSWER

        use_sentence_answers = self.prompt_strategy.is_split_proposer_prompt() or self.prompt_strategy.is_moa_proposer_prompt()
        input_modality_to_example_answer = {
            Modality.TEXT: TEXT_SENTENCE_EXAMPLE_ANSWER if use_sentence_answers else TEXT_EXAMPLE_ANSWER,
            Modality.IMAGE: IMAGE_SENTENCE_EXAMPLE_ANSWER if use_sentence_answers else IMAGE_EXAMPLE_ANSWER,
            Modality.AUDIO: AUDIO_SENTENCE_EXAMPLE_ANSWER if use_sentence_answers else AUDIO_EXAMPLE_ANSWER,
        }

        example_answer = ""
        for input_modality in input_modalities:
            example_answer += input_modality_to_example_answer[input_modality].rstrip()
            if use_sentence_answers:
                example_answer += " "
        example_answer = example_answer + "\n"

        return example_answer

    def _get_all_format_kwargs(
        self,
        candidate: DataRecord | list[DataRecord],
        input_fields: list[str],
        input_modalities: set[Modality],
        output_fields: list[str],
        right_candidate: DataRecord | None,
        right_input_fields: list[str],
        right_input_modalities: set[Modality],
        **kwargs,
    ) -> dict:
        """
        Returns a dictionary containing all the format kwargs for templating the prompts.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.
            output_fields (list[str]): The output fields.
            kwargs: The keyword arguments provided by the user.

        Returns:
            dict: The dictionary containing all the format kwargs.
        """
        # get format kwargs which depend on the input data
        input_format_kwargs = {
            "context": self._get_context(candidate, input_fields),
            "input_fields_desc": self._get_input_fields_desc(candidate[0] if isinstance(candidate, list) else candidate, input_fields),
            "output_fields_desc": self._get_output_fields_desc(output_fields, **kwargs),
            "agg_instruction": self._get_agg_instruction(**kwargs),
            "filter_condition": self._get_filter_condition(**kwargs),
            "join_condition": self._get_join_condition(**kwargs),
            "original_output": self._get_original_output(**kwargs),
            "critique_output": self._get_critique_output(**kwargs),
            "model_responses": self._get_model_responses(**kwargs),
            "chunk_outputs": self._get_chunk_outputs(**kwargs),
        }

        # if a right candidate is provided, we also get the context and input field descriptions for the right candidate
        if right_candidate is not None:
            input_format_kwargs.update({
                "right_context": self._get_context(right_candidate, right_input_fields),
                "right_input_fields_desc": self._get_input_fields_desc(right_candidate, right_input_fields),
            })

        # get format kwargs which depend on the prompt strategy
        full_input_modalities = input_modalities.union(right_input_modalities)
        prompt_strategy_format_kwargs = {
            "output_format_instruction": self._get_output_format_instruction(),
            "job_instruction": self._get_job_instruction(full_input_modalities),
            "desc_section": self._get_desc_section(),
            "critique_criteria": self._get_critique_criteria(),
            "refinement_criteria": self._get_refinement_criteria(),
            "finish_instruction": self._get_finish_instruction(),
            "example_input_fields": self._get_example_input_fields(input_modalities),
            "right_example_input_fields": self._get_example_input_fields(right_input_modalities, right=True),
            "example_output_fields": self._get_example_output_fields(input_modalities),
            "example_context": self._get_example_context(input_modalities),
            "second_example_context": self._get_example_context(input_modalities, second=True) if self.prompt_strategy.is_agg_prompt() else "",
            "third_example_context": self._get_example_context(input_modalities, third=True) if self.prompt_strategy.is_agg_prompt() else "",
            "right_example_context": self._get_example_context(right_input_modalities, right=True),
            "image_disclaimer": self._get_image_disclaimer(input_modalities, agg=self.prompt_strategy.is_agg_prompt()),
            "audio_disclaimer": self._get_audio_disclaimer(input_modalities, agg=self.prompt_strategy.is_agg_prompt()),
            "right_image_disclaimer": self._get_image_disclaimer(right_input_modalities, right=True),
            "right_audio_disclaimer": self._get_audio_disclaimer(right_input_modalities, right=True),
            "example_agg_instruction": EXAMPLE_AGG_INSTRUCTION,
            "example_filter_condition": EXAMPLE_FILTER_CONDITION,
            "example_join_condition": EXAMPLE_JOIN_CONDITION,
            "example_reasoning": self._get_example_reasoning(input_modalities),
            "example_answer": self._get_example_answer(input_modalities),
        }

        # return all format kwargs
        return {**input_format_kwargs, **prompt_strategy_format_kwargs}

    def _create_audio_messages(self, candidate: DataRecord | list[DataRecord], input_fields: list[str]) -> list[dict]:
        """
        Parses the candidate record(s) and returns the audio messages for the chat payload.

        Args:
            candidate (DataRecord | list[DataRecord]): The input record(s).
            input_fields (list[str]): The list of input fields.

        Returns:
            list[dict]: The audio messages for the chat payload.
        """
        # normalize type to be list[DataRecord]
        if isinstance(candidate, DataRecord):
            candidate = [candidate]

        # create a message for each audio recording in an input field with an audio (or list of audio) type
        audio_content = []
        for field_name in input_fields:
            for dr in candidate:
                field_value = dr[field_name]
                field_type = dr.get_field_type(field_name)

                # audio filepath (or list of audio filepaths)
                if field_type.annotation in [AudioFilepath, AudioFilepath | None, AudioFilepath | Any] and field_value is not None:
                    with open(field_value, "rb") as f:
                        base64_audio_str = base64.b64encode(f.read()).decode("utf-8")
                    audio_content.append(
                        {"type": "input_audio", "input_audio": {"data": base64_audio_str, "format": "wav"}}
                    )

                elif field_type.annotation in [list[AudioFilepath], list[AudioFilepath] | None, list[AudioFilepath] | Any]:
                    for audio_filepath in field_value:
                        if audio_filepath is None:
                            continue
                        with open(audio_filepath, "rb") as f:
                            base64_audio_str = base64.b64encode(f.read()).decode("utf-8")
                        audio_content.append(
                            {"type": "input_audio", "input_audio": {"data": base64_audio_str, "format": "wav"}}
                        )

                # pre-encoded images (or list of pre-encoded images)
                elif field_type.annotation in [AudioBase64, AudioBase64 | None, AudioBase64 | Any] and field_value is not None:
                    audio_content.append(
                        {"type": "input_audio", "input_audio": {"data": field_value, "format": "wav"}}
                    )

                elif field_type.annotation in [list[AudioBase64], list[AudioBase64] | None, list[AudioBase64] | Any]:
                    for base64_audio in field_value:
                        if base64_audio is None:
                            continue
                        audio_content.append(
                            {"type": "input_audio", "input_audio": {"data": base64_audio, "format": "wav"}}
                        )

        return [{"role": "user", "type": "input_audio", "content": audio_content}] if len(audio_content) > 0 else []

    def _create_image_messages(self, candidate: DataRecord | list[DataRecord], input_fields: list[str]) -> list[dict]:
        """
        Parses the candidate record(s) and returns the image messages for the chat payload.

        Args:
            candidate (DataRecord | list[DataRecord]): The input record(s).
            input_fields (list[str]): The list of input fields.

        Returns:
            list[dict]: The image messages for the chat payload.
        """
        # normalize type to be list[DataRecord]
        if isinstance(candidate, DataRecord):
            candidate = [candidate]

        # create a message for each image in an input field with an image (or list of image) type
        image_content = []
        for field_name in input_fields:
            for dr in candidate:
                field_value = dr[field_name]
                field_type = dr.get_field_type(field_name)

                # image filepath (or list of image filepaths)
                if field_type.annotation in [ImageFilepath, ImageFilepath | None, ImageFilepath | Any] and field_value is not None:
                    with open(field_value, "rb") as f:
                        base64_image_str = base64.b64encode(f.read()).decode("utf-8")
                    image_content.append(
                        {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image_str}"}}
                    )

                elif field_type.annotation in [list[ImageFilepath], list[ImageFilepath] | None, list[ImageFilepath] | Any]:
                    for image_filepath in field_value:
                        if image_filepath is None:
                            continue
                        with open(image_filepath, "rb") as f:
                            base64_image_str = base64.b64encode(f.read()).decode("utf-8")
                        image_content.append(
                            {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image_str}"}}
                        )

                # image url (or list of image urls)
                elif field_type.annotation in [ImageURL, ImageURL | None, ImageURL | Any] and field_value is not None:
                    image_content.append({"type": "image_url", "image_url": {"url": field_value}})

                elif field_type.annotation in [list[ImageURL], list[ImageURL] | None, list[ImageURL] | Any]:
                    for image_url in field_value:
                        if image_url is None:
                            continue
                        image_content.append({"type": "image_url", "image_url": {"url": image_url}})

                # pre-encoded images (or list of pre-encoded images)
                elif field_type.annotation in [ImageBase64, ImageBase64 | None, ImageBase64 | Any] and field_value is not None:
                    image_content.append(
                        {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{field_value}"}}
                    )

                elif field_type.annotation in [list[ImageBase64], list[ImageBase64] | None, list[ImageBase64] | Any]:
                    for base64_image in field_value:
                        if base64_image is None:
                            continue
                        image_content.append(
                            {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"}}
                        )

        return [{"role": "user", "type": "image", "content": image_content}] if len(image_content) > 0 else []

    def _get_system_prompt(self, **format_kwargs) -> str | None:
        """
        Returns the fully templated system prompt for the given prompt strategy.
        Returns None if the prompt strategy does not use a system prompt.

        Returns:
            str | None: The fully templated system prompt (or None if the prompt strategy
                does not use a system prompt).
        """
        base_prompt: str = self.BASE_SYSTEM_PROMPT_MAP.get(self.prompt_strategy)

        # for critic and refine prompt strategies, we do not use a base prompt
        if base_prompt is None:
            return base_prompt

        return base_prompt.format(**format_kwargs)

    def _get_user_messages(self, candidate: DataRecord | list[DataRecord], input_fields: list[str], right_candidate: DataRecord | None, right_input_fields: list[str], **kwargs) -> str:
        """
        Returns a list of messages for the chat payload based on the prompt strategy.

        Args:
            candidate (DataRecord | list[DataRecord]): The input record(s).
            input_fields (list[str]): The input fields.
            output_fields (list[str]): The output fields.
            kwargs: The formatting kwargs and some keyword arguments provided by the user.

        Returns:
            Tuple[str, str | None]: The fully templated start and end of the user prompt.
                The second element will be None for text prompts.
        """
        # get the base prompt template
        base_prompt = self.BASE_USER_PROMPT_MAP.get(self.prompt_strategy)

        # get any image messages for the chat payload (will be an empty list if no image fields exist)
        image_messages = self._create_image_messages(candidate, input_fields)

        # get any audio messages for the chat payload (will be an empty list if no audio fields exist)
        audio_messages = self._create_audio_messages(candidate, input_fields)

        # get any right image / audio messages for the chat payload (will be an empty list if image / audio not present)
        right_image_messages, right_audio_messages = [], []
        if self.prompt_strategy.is_join_prompt():
            assert right_candidate is not None, "Right candidate must be provided for join prompts."
            right_image_messages = self._create_image_messages(right_candidate, right_input_fields)
            right_audio_messages = self._create_audio_messages(right_candidate, right_input_fields)

        # get any original messages for critique and refinement operations
        original_messages = kwargs.get("original_messages")
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            assert original_messages is not None, (
                "Original messages must be provided for critique and refinement operations."
            )

        # combine image and audio messages
        image_audio_messages = image_messages + audio_messages
        right_image_audio_messages = right_image_messages + right_audio_messages
        has_image_audio = len(image_audio_messages) > 0
        has_right_image_audio = len(right_image_audio_messages) > 0

        # construct the user messages based on the prompt strategy
        user_messages = []
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            # NOTE: if this critic / refinement prompt is processing images / audio, those images / audio
            # will be part of the `original_messages` and will show up in the final chat payload
            base_prompt_start, base_prompt_end = base_prompt.split("<<original-prompt-placeholder>>\n")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(original_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        # handle joins with left and right images / audio
        elif self.prompt_strategy.is_join_prompt() and has_image_audio and has_right_image_audio:
            base_prompt_start, base_prompt_rest = base_prompt.split("<<image-audio-placeholder>>")
            base_prompt_mid, base_prompt_end = base_prompt_rest.split("<<right-image-audio-placeholder>>")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(image_audio_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_mid.format(**kwargs)})
            user_messages.extend(right_image_audio_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        # handle joins with only left images / audio
        elif self.prompt_strategy.is_join_prompt() and has_image_audio and not has_right_image_audio:
            base_prompt = base_prompt.replace("<<right-image-audio-placeholder>>", "")
            base_prompt_start, base_prompt_end = base_prompt.split("<<image-audio-placeholder>>")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(image_audio_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        # handle joins with only right images / audio
        elif self.prompt_strategy.is_join_prompt() and not has_image_audio and has_right_image_audio:
            base_prompt = base_prompt.replace("<<image-audio-placeholder>>", "")
            base_prompt_start, base_prompt_end = base_prompt.split("<<right-image-audio-placeholder>>")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(right_image_audio_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        # handle non-joins with images / audio
        elif not self.prompt_strategy.is_join_prompt() and has_image_audio and not self.prompt_strategy.is_moa_aggregator_prompt():
            base_prompt_start, base_prompt_end = base_prompt.split("<<image-audio-placeholder>>")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(image_audio_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        # handle prompts w/no images or audio
        else:
            base_prompt = base_prompt.replace("<<image-audio-placeholder>>", "")
            base_prompt = base_prompt.replace("<<right-image-audio-placeholder>>", "")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt.format(**kwargs)})

        return user_messages

    def create_messages(self, candidate: DataRecord | list[DataRecord], output_fields: list[str], right_candidate: DataRecord | None = None, **kwargs) -> list[dict]:
        """
        Creates the messages for the chat payload based on the prompt strategy.

        Each message will be a dictionary with the following format:
        {
            "role": "user" | "system",
            "type": "text" | "image",
            "content": str
        }

        Args:
            candidate (DataRecord | list[DataRecord]): The input record(s).
            output_fields (list[str]): The output fields.
            right_candidate (DataRecord | None): The other join input record (only provided for joins).
            kwargs: The keyword arguments provided by the user.

        Returns:
            list[dict]: The messages for the chat payload.
        """
        # compute the set of input fields
        input_fields = self._get_input_fields(candidate[0] if isinstance(candidate, list) else candidate, **kwargs)
        right_input_fields = [] if right_candidate is None else self._get_input_fields(right_candidate, **kwargs)

        # use input fields to determine the left / right input modalities
        input_modalities = self._get_input_modalities(candidate[0] if isinstance(candidate, list) else candidate, input_fields)
        right_input_modalities = set() if right_candidate is None else self._get_input_modalities(right_candidate, right_input_fields)

        # initialize messages
        messages = []

        # compute the full dictionary of format kwargs and add to kwargs
        format_kwargs = self._get_all_format_kwargs(candidate, input_fields, input_modalities, output_fields, right_candidate, right_input_fields, right_input_modalities, **kwargs)
        kwargs = {**kwargs, **format_kwargs}

        # generate system message (if applicable)
        system_prompt = self._get_system_prompt(**kwargs)
        if system_prompt is not None:
            messages.append({"role": "system", "type": "text", "content": system_prompt})

        # generate user messages and add to messages
        user_messages = self._get_user_messages(candidate, input_fields, right_candidate, right_input_fields, **kwargs)
        messages.extend(user_messages)

        return messages
